# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import copy

import numpy as np
import torch
from mmcv import ConfigDict, build_from_cfg

from otx.algorithms.classification.adapters.mmcls.utils.builder import build_classifier
from otx.mpa.stage import Stage
from otx.mpa.utils.config_utils import recursively_update_cfg, update_or_add_custom_hook
from otx.mpa.utils.logger import get_logger

logger = get_logger()


class ClsStage(Stage):
    MODEL_BUILDER = build_classifier

    def configure(self, model_cfg, model_ckpt, data_cfg, training=True, **kwargs):  # noqa: C901
        """Create MMCV-consumable config from given inputs"""
        logger.info(f"configure: training={training}")

        # Recipe + model
        cfg = self.cfg
        self.configure_model(cfg, model_cfg, training, **kwargs)
        self.configure_ckpt(cfg, model_ckpt, kwargs.get("pretrained", None))
        self.configure_data(cfg, data_cfg, training, **kwargs)
        self.configure_task(cfg, training, **kwargs)
        return cfg

    def configure_model(self, cfg, model_cfg, training, **kwargs):
        if model_cfg:
            if hasattr(cfg, "model"):
                cfg.merge_from_dict(model_cfg._cfg_dict)
            else:
                cfg.model = copy.deepcopy(model_cfg.model)

        cfg.model_task = cfg.model.pop("task", "classification")
        if cfg.model_task != "classification":
            raise ValueError(f"Given model_cfg ({model_cfg.filename}) is not supported by classification recipe")

        # OV-plugin
        ir_model_path = kwargs.get("ir_model_path")
        if ir_model_path:

            def is_mmov_model(k, v):
                if k == "type" and v.startswith("MMOV"):
                    return True
                return False

            ir_weight_path = kwargs.get("ir_weight_path", None)
            ir_weight_init = kwargs.get("ir_weight_init", False)
            recursively_update_cfg(
                cfg,
                is_mmov_model,
                {"model_path": ir_model_path, "weight_path": ir_weight_path, "init_weight": ir_weight_init},
            )
        self.configure_in_channel(cfg)
        self.configure_topk(cfg)

    @staticmethod
    def configure_in_channel(cfg):
        configure_required = False
        if cfg.model.get("neck") is not None:
            if cfg.model.neck.get("in_channels") is not None and cfg.model.neck.in_channels <= 0:
                configure_required = True
        if not configure_required and cfg.model.get("head") is not None:
            if cfg.model.head.get("in_channels") is not None and cfg.model.head.in_channels <= 0:
                configure_required = True
        if not configure_required:
            return

        # update model layer's in/out configuration
        from mmcv.cnn import MODELS as backbone_reg

        layer = build_from_cfg(cfg.model.backbone, backbone_reg)
        layer.eval()
        input_shape = [3, 224, 224]
        # MMOV model
        if hasattr(layer, "input_shapes"):
            input_shape = next(iter(getattr(layer, "input_shapes").values()))
            input_shape = input_shape[1:]
            if any(i < 0 for i in input_shape):
                input_shape = [3, 244, 244]
        logger.debug(f"input shape for backbone {input_shape}")
        output = layer(torch.rand([1] + list(input_shape)))
        if isinstance(output, (tuple, list)):
            output = output[-1]
        in_channels = output.shape[1]
        if cfg.model.get("neck") is not None:
            if cfg.model.neck.get("in_channels") is not None:
                logger.info(
                    f"'in_channels' config in model.neck is updated from "
                    f"{cfg.model.neck.in_channels} to {in_channels}"
                )
                cfg.model.neck.in_channels = in_channels
                logger.debug(f"input shape for neck {input_shape}")
                from mmcls.models.builder import NECKS as neck_reg

                layer = build_from_cfg(cfg.model.neck, neck_reg)
                layer.eval()
                output = layer(torch.rand(output.shape))
                if isinstance(output, (tuple, list)):
                    output = output[-1]
                in_channels = output.shape[1]
        if cfg.model.get("head") is not None:
            if cfg.model.head.get("in_channels") is not None:
                logger.info(
                    f"'in_channels' config in model.head is updated from "
                    f"{cfg.model.head.in_channels} to {in_channels}"
                )
                cfg.model.head.in_channels = in_channels

    @staticmethod
    def configure_topk(cfg):
        if cfg.model.head.get("topk", False) and isinstance(cfg.model.head.topk, tuple):
            cfg.model.head.topk = (1,) if cfg.model.head.num_classes < 5 else (1, 5)
            if cfg.model.get("multilabel", False) or cfg.model.get("hierarchical", False):
                cfg.model.head.pop("topk", None)

    def configure_ckpt(self, cfg, model_ckpt, pretrained):
        # Checkpoint
        if model_ckpt:
            cfg.load_from = self.get_model_ckpt(model_ckpt)
        if pretrained and isinstance(pretrained, str):
            logger.info(f"Overriding cfg.load_from -> {pretrained}")
            cfg.load_from = pretrained  # Overriding by stage input
        if cfg.get("resume", False):
            cfg.resume_from = cfg.load_from
        if cfg.get("load_from", None) and cfg.model.backbone.get("pretrained", None):
            cfg.model.backbone.pretrained = None

    def configure_data(self, cfg, data_cfg, training, **kwargs):
        # Data
        if data_cfg:
            cfg.merge_from_dict(data_cfg)
        Stage.configure_data(cfg, training, **kwargs)

    def configure_task(self, cfg, training, **kwargs):
        self.configure_classes(cfg)

    def configure_classes(self, cfg):
        model_classes, data_classes = [], []
        self.model_meta = self.get_model_meta(cfg)
        train_data_cfg = Stage.get_data_cfg(cfg, "train")
        if isinstance(train_data_cfg, list):
            train_data_cfg = train_data_cfg[0]

        model_classes = Stage.get_model_classes(cfg)
        data_classes = Stage.get_data_classes(cfg)

        if cfg.get("model_classes", []):
            cfg.model.head.num_classes = len(cfg.model_classes)
        elif model_classes:
            cfg.model.head.num_classes = len(model_classes)
        elif data_classes:
            cfg.model.head.num_classes = len(data_classes)
        self.model_meta["CLASSES"] = model_classes

        if not train_data_cfg.get("new_classes", False):  # when train_data_cfg doesn't have 'new_classes' key
            new_classes = np.setdiff1d(data_classes, model_classes).tolist()
            train_data_cfg["new_classes"] = new_classes
        self.model_classes = model_classes
        self.data_classes = data_classes



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from collections.abc import MutableMapping

import numpy as np
import torch


def sync_batchnorm_2_batchnorm(module, dim=2):
    if dim == 1:
        bn = torch.nn.BatchNorm1d
    elif dim == 2:
        bn = torch.nn.BatchNorm2d
    elif dim == 3:
        bn = torch.nn.BatchNorm3d
    else:
        raise NotImplementedError()

    module_output = module
    if isinstance(module, torch.nn.SyncBatchNorm):
        module_output = bn(
            module.num_features,
            module.eps,
            module.momentum,
            module.affine,
            module.track_running_stats,
        )
        if module.affine:
            module_output.weight.data = module.weight.data.clone().detach()
            module_output.bias.data = module.bias.data.clone().detach()
            module_output.weight.requires_grad = module.weight.requires_grad
            module_output.bias.requires_grad = module.bias.requires_grad

        module_output.running_mean = module.running_mean
        module_output.running_var = module.running_var
        module_output.num_batches_tracked = module.num_batches_tracked
        if hasattr(module, "qconfig"):
            module_output.qconfig = module.qconfig

    for name, child in module.named_children():
        module_output.add_module(name, sync_batchnorm_2_batchnorm(child, dim))

    del module

    return module_output


def numpy_2_list(data):

    if isinstance(data, np.ndarray):
        return data.tolist()

    if isinstance(data, MutableMapping):
        for key, value in data.items():
            data[key] = numpy_2_list(value)
    elif isinstance(data, (list, tuple)):
        data_ = []
        for value in data:
            data_.append(numpy_2_list(value))
        if isinstance(data, tuple):
            data = tuple(data_)
        else:
            data = data_
    return data



"""Unit Test for otx.algorithms.action.tasks.openvino."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os
from typing import Any, Dict

import numpy as np
import pytest
from openvino.model_zoo.model_api.adapters import OpenvinoAdapter

from otx.algorithms.action.adapters.openvino import ActionOVClsDataLoader
from otx.algorithms.action.configs.base.configuration import ActionConfig
from otx.algorithms.action.tasks.openvino import (
    ActionOpenVINOInferencer,
    ActionOpenVINOTask,
    DataLoaderWrapper,
)
from otx.api.configuration import ConfigurableParameters
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.label import Domain
from otx.api.entities.label_schema import LabelGroup, LabelGroupType, LabelSchemaEntity
from otx.api.entities.model import (
    ModelConfiguration,
    ModelEntity,
    ModelFormat,
    ModelOptimizationType,
    ModelPrecision,
    OptimizationMethod,
)
from otx.api.entities.model_template import InstantiationType, TaskFamily, TaskType
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.exportable_code.inference import BaseInferencer
from otx.api.usecases.exportable_code.prediction_to_annotation_converter import (
    ClassificationToAnnotationConverter,
    DetectionBoxToAnnotationConverter,
)
from otx.api.usecases.tasks.interfaces.optimization_interface import OptimizationType
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.action.test_helpers import (
    MockModelTemplate,
    generate_action_cls_otx_dataset,
    generate_labels,
    return_args,
)


class MockOVInferencer(BaseInferencer):
    """Mock class for OV inferencer."""

    def __init__(self, *args, **kwargs):
        self.model = MockModel()
        self.model.t = 8
        self.model.w = 256
        self.model.h = 256
        self.labels = generate_labels(1, Domain.ACTION_CLASSIFICATION)
        self.configuration: Dict[Any, Any] = {}

    def predict(self, data):
        return AnnotationSceneEntity(
            annotations=[Annotation(shape=Rectangle(0, 0, 1, 1), labels=[ScoredLabel(self.labels[0], 1.0)])],
            kind=AnnotationSceneKind.PREDICTION,
        )

    def pre_process(self, item):
        return item, {"dummy_meta": "dummy_info"}

    def forward(self, item):
        pass

    def post_process(self, item):
        pass


class MockModel:
    """Mock class for OV model."""

    def preprocess(self, image):
        return "Preprocess function is called"

    def postprocess(self, prediction, metadata):
        return "Postprocess function is called"

    def infer_sync(self, image):
        return "Funtion infer_sync is called"


class MockOpenvinoAdapter(OpenvinoAdapter):
    """Mock class for OpenvinoAdapter."""

    def __init__(self, *args, **kwargs):
        pass


class MockDataloader(ActionOVClsDataLoader):
    """Mock class for dataloader for OpenVINO inference."""

    def __init__(self, dataset, *args, **kwargs):
        self.dataset = dataset

    def __len__(self):
        return 1

    def __getitem__(self, index):
        if index >= len(self):
            raise StopIteration
        return self.dataset._items

    def add_prediction(self, dataset, data, prediction):
        for dataset_item in dataset:
            dataset_item.append_labels(prediction.annotations[0].get_labels())


class TestActionOVInferencer:
    """Test class for ActionOpenVINOInferencer."""

    @pytest.fixture(autouse=True)
    def setup(self, mocker) -> None:
        self.labels = generate_labels(3, Domain.ACTION_CLASSIFICATION)
        self.label_schema = LabelSchemaEntity()
        label_group = LabelGroup(
            name="labels",
            labels=self.labels,
            group_type=LabelGroupType.EXCLUSIVE,
        )
        self.label_schema.add_group(label_group)
        mocker.patch("otx.algorithms.action.tasks.openvino.OpenvinoAdapter.__init__", return_value=None)
        mocker.patch("otx.algorithms.action.tasks.openvino.Model.create_model", return_value=MockModel())
        self.inferencer = ActionOpenVINOInferencer(
            "ACTION_CLASSIFICATION",
            ActionConfig(),
            self.label_schema,
            "openvino.xml",
            "openvino.bin",
        )

    @e2e_pytest_unit
    def test_init(self) -> None:
        """Test __init__ function."""
        inferencer = ActionOpenVINOInferencer(
            "ACTION_CLASSIFICATION",
            ActionConfig(),
            self.label_schema,
            "openvino.xml",
            "openvino.bin",
        )
        assert inferencer.task_type == "ACTION_CLASSIFICATION"
        assert inferencer.label_schema == self.label_schema
        assert isinstance(inferencer.model, MockModel)
        assert isinstance(inferencer.converter, ClassificationToAnnotationConverter)

        inferencer = ActionOpenVINOInferencer(
            "ACTION_DETECTION",
            ActionConfig(),
            self.label_schema,
            "openvino.xml",
            "openvino.bin",
        )
        assert inferencer.task_type == "ACTION_DETECTION"
        assert isinstance(inferencer.converter, DetectionBoxToAnnotationConverter)

    @e2e_pytest_unit
    def test_pre_process(self) -> None:
        """Test pre_process funciton."""
        dataset = generate_action_cls_otx_dataset(1, 10, self.labels)
        inputs = dataset._items
        assert self.inferencer.pre_process(inputs) == "Preprocess function is called"

    @e2e_pytest_unit
    def test_post_process(self, mocker) -> None:
        """Test post_process function."""
        mocker.patch(
            "otx.algorithms.action.tasks.openvino.ClassificationToAnnotationConverter.convert_to_annotation",
            side_effect=return_args,
        )
        assert (
            self.inferencer.post_process({"dummy": np.ndarray(1)}, {"dummy": "meta"})[0][0]
            == "Postprocess function is called"
        )

    @e2e_pytest_unit
    def test_forward(self) -> None:
        """Test forward function."""

        assert self.inferencer.forward({"dummy": np.ndarray(1)}) == "Funtion infer_sync is called"

    @e2e_pytest_unit
    def test_predict(self, mocker) -> None:
        """Test predict function."""

        mocker.patch(
            "otx.algorithms.action.tasks.openvino.ActionOpenVINOInferencer.pre_process",
            return_value=("data", "metadata"),
        )
        mocker.patch(
            "otx.algorithms.action.tasks.openvino.ActionOpenVINOInferencer.forward", return_value="raw_predictions"
        )
        mocker.patch(
            "otx.algorithms.action.tasks.openvino.ActionOpenVINOInferencer.post_process", return_value="predictions"
        )

        dataset = generate_action_cls_otx_dataset(1, 10, self.labels)
        inputs = dataset._items
        assert self.inferencer.predict(inputs) == "predictions"


class TestActionOVTask:
    """Test class for ActionOpenVINOTask."""

    @pytest.fixture(autouse=True)
    def setup(self, mocker) -> None:
        self.video_len = 1
        self.frame_len = 10

        labels = generate_labels(3, Domain.ACTION_CLASSIFICATION)
        self.label_schema = LabelSchemaEntity()
        label_group = LabelGroup(
            name="labels",
            labels=labels,
            group_type=LabelGroupType.EXCLUSIVE,
        )
        self.label_schema.add_group(label_group)
        template = MockModelTemplate(
            model_template_id="template_id",
            model_template_path="template_path",
            name="template",
            task_family=TaskFamily.VISION,
            task_type=TaskType.ACTION_CLASSIFICATION,
            instantiation=InstantiationType.CLASS,
        )

        config = ModelConfiguration(ActionConfig(), self.label_schema)
        self.dataset = generate_action_cls_otx_dataset(1, 10, labels)
        self.model = ModelEntity(self.dataset, config)
        self.model.set_data("openvino.xml", np.ndarray([1]).tobytes())
        self.model.set_data("openvino.bin", np.ndarray([1]).tobytes())

        self.task_environment = TaskEnvironment(
            model=self.model,
            hyper_parameters=ConfigurableParameters(header="h-params"),
            label_schema=self.label_schema,
            model_template=template,
        )

    @e2e_pytest_unit
    def test_load_inferencer(self, mocker) -> None:
        """Test load_inferencer function."""

        mocker.patch("otx.algorithms.action.tasks.openvino.ActionOpenVINOInferencer", return_value=MockOVInferencer())
        task = ActionOpenVINOTask(self.task_environment)
        assert isinstance(task.inferencer, MockOVInferencer)

        self.task_environment.model = None
        with pytest.raises(RuntimeError):
            task = ActionOpenVINOTask(self.task_environment)

    @e2e_pytest_unit
    def test_infer(self, mocker) -> None:
        """Test infer function."""

        mocker.patch(
            "otx.algorithms.action.tasks.openvino.ActionOpenVINOTask.load_inferencer", return_value=MockOVInferencer()
        )
        mocker.patch("otx.algorithms.action.tasks.openvino.get_ovdataloader", return_value=MockDataloader(self.dataset))
        task = ActionOpenVINOTask(self.task_environment)
        output = task.infer(self.dataset.with_empty_annotations())
        assert output[0].annotation_scene.kind == AnnotationSceneKind.PREDICTION

    @e2e_pytest_unit
    def test_evaluate(self, mocker) -> None:
        """Test evaluate function."""

        class MockPerformance:
            def get_performance(self):
                return 1.0

        mocker.patch(
            "otx.algorithms.action.tasks.openvino.ActionOpenVINOTask.load_inferencer", return_value=MockOVInferencer()
        )
        task = ActionOpenVINOTask(self.task_environment)

        resultset = ResultSetEntity(
            self.model,
            self.dataset,
            self.dataset,
        )

        mocker.patch.object(MetricsHelper, "compute_accuracy", return_value=MockPerformance())
        task.evaluate(resultset, "Accuracy")
        assert resultset.performance == 1.0

        mocker.patch.object(MetricsHelper, "compute_f_measure", return_value=MockPerformance())
        self.task_environment.model_template.task_type = TaskType.ACTION_DETECTION
        task = ActionOpenVINOTask(self.task_environment)
        task.evaluate(resultset, "Accuracy")
        assert resultset.performance == 1.0

    @e2e_pytest_unit
    def test_deploy(self, mocker) -> None:
        """Test function for deploy function."""

        mocker.patch(
            "otx.algorithms.action.tasks.openvino.ActionOpenVINOTask.load_inferencer", return_value=MockOVInferencer()
        )
        task = ActionOpenVINOTask(self.task_environment)
        assert self.model.exportable_code is None
        task.deploy(self.model)
        assert self.model.exportable_code is not None

    @e2e_pytest_unit
    def test_optimize(self, mocker) -> None:
        """Test optimization function."""

        class MockPipeline:
            """Mock class for POT pipeline"""

            def run(self, model):
                return model

        def mock_save_model(model, tempdir, model_name):
            """Mock function for save_model function."""
            with open(os.path.join(tempdir, "model.xml"), "wb") as f:
                f.write(np.ndarray(1).tobytes())
            with open(os.path.join(tempdir, "model.bin"), "wb") as f:
                f.write(np.ndarray(1).tobytes())

        mocker.patch("otx.algorithms.action.tasks.openvino.get_ovdataloader", return_value=MockDataloader(self.dataset))
        mocker.patch(
            "otx.algorithms.action.tasks.openvino.DataLoaderWrapper", return_value=MockDataloader(self.dataset)
        )
        mocker.patch("otx.algorithms.action.tasks.openvino.load_model", return_value=self.model)
        mocker.patch("otx.algorithms.action.tasks.openvino.get_nodes_by_type", return_value=False)
        mocker.patch("otx.algorithms.action.tasks.openvino.IEEngine", return_value=True)
        mocker.patch("otx.algorithms.action.tasks.openvino.create_pipeline", return_value=MockPipeline())
        mocker.patch("otx.algorithms.action.tasks.openvino.compress_model_weights", return_value=True)
        mocker.patch("otx.algorithms.action.tasks.openvino.save_model", side_effect=mock_save_model)
        mocker.patch(
            "otx.algorithms.action.tasks.openvino.ActionOpenVINOTask.load_inferencer", return_value=MockOVInferencer()
        )
        task = ActionOpenVINOTask(self.task_environment)
        task.optimize(OptimizationType.POT, self.dataset, self.model, OptimizationParameters())
        assert self.model.get_data("openvino.xml") is not None
        assert self.model.get_data("openvino.bin") is not None
        assert self.model.model_format == ModelFormat.OPENVINO
        assert self.model.optimization_type == ModelOptimizationType.POT
        assert self.model.optimization_methods == [OptimizationMethod.QUANTIZATION]
        assert self.model.precision == [ModelPrecision.INT8]


class TestDataLoaderWrapper:
    """Test class for DataLoaderWrapper"""

    def setup(self, mocker) -> None:
        labels = generate_labels(3, Domain.ACTION_CLASSIFICATION)
        self.dataset = generate_action_cls_otx_dataset(1, 10, labels)
        ovdataloader = MockDataloader(self.dataset)
        inferencer = MockOVInferencer()
        self.dataloader = DataLoaderWrapper(ovdataloader, inferencer)

    @e2e_pytest_unit
    def test_len(self) -> None:
        """Test __len__ function."""

        assert len(self.dataloader) == 1

    def test_getitem(self) -> None:
        """Test __getitem__ function."""

        out = self.dataloader[0]
        assert out[0][0] == 0
        assert isinstance(out[0][1], AnnotationSceneEntity)
        assert len(out[1]) == 10
        assert isinstance(out[2], dict)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
import torch.nn as nn
import torch.nn.functional as F
from mmcls.models.builder import HEADS
from mmcls.models.heads import MultiLabelClsHead
from mmcv.cnn import normal_init


@HEADS.register_module()
class CustomMultiLabelLinearClsHead(MultiLabelClsHead):
    """Custom Linear classification head for multilabel task.
    Args:
        num_classes (int): Number of categories.
        in_channels (int): Number of channels in the input feature map.
        normalized (bool): Normalize input features and weights.
        scale (float): positive scale parameter.
        loss (dict): Config of classification loss.
    """

    def __init__(
        self,
        num_classes,
        in_channels,
        normalized=False,
        scale=1.0,
        loss=dict(type="CrossEntropyLoss", use_sigmoid=True, reduction="mean", loss_weight=1.0),
    ):
        super(CustomMultiLabelLinearClsHead, self).__init__(loss=loss)
        if num_classes <= 0:
            raise ValueError(f"num_classes={num_classes} must be a positive integer")

        self.in_channels = in_channels
        self.num_classes = num_classes
        self.normalized = normalized
        self.scale = scale
        self._init_layers()

    def _init_layers(self):
        if self.normalized:
            self.fc = AnglularLinear(self.in_channels, self.num_classes)
        else:
            self.fc = nn.Linear(self.in_channels, self.num_classes)

    def init_weights(self):
        if isinstance(self.fc, nn.Linear):
            normal_init(self.fc, mean=0, std=0.01, bias=0)

    def loss(self, cls_score, gt_label, valid_label_mask=None):
        gt_label = gt_label.type_as(cls_score)
        num_samples = len(cls_score)
        losses = dict()

        # map difficult examples to positive ones
        _gt_label = torch.abs(gt_label)
        # compute loss
        loss = self.compute_loss(cls_score, _gt_label, valid_label_mask=valid_label_mask, avg_factor=num_samples)
        losses["loss"] = loss / self.scale
        return losses

    def forward_train(self, x, gt_label, **kwargs):
        img_metas = kwargs.get("img_metas", False)
        gt_label = gt_label.type_as(x)
        cls_score = self.fc(x) * self.scale

        valid_batch_mask = gt_label >= 0
        gt_label = gt_label[
            valid_batch_mask,
        ].view(gt_label.shape[0], -1)
        cls_score = cls_score[
            valid_batch_mask,
        ].view(cls_score.shape[0], -1)
        if img_metas:
            valid_label_mask = self.get_valid_label_mask(img_metas=img_metas)
            valid_label_mask = valid_label_mask.to(x.device)
            valid_label_mask = valid_label_mask[
                valid_batch_mask,
            ].view(valid_label_mask.shape[0], -1)
            losses = self.loss(cls_score, gt_label, valid_label_mask=valid_label_mask)
        else:
            losses = self.loss(cls_score, gt_label)
        return losses

    def simple_test(self, img):
        """Test without augmentation."""
        cls_score = self.fc(img) * self.scale
        if isinstance(cls_score, list):
            cls_score = sum(cls_score) / float(len(cls_score))
        if torch.onnx.is_in_onnx_export():
            return cls_score
        pred = torch.sigmoid(cls_score) if cls_score is not None else None
        pred = list(pred.detach().cpu().numpy())
        return pred

    def get_valid_label_mask(self, img_metas):
        valid_label_mask = []
        for i, meta in enumerate(img_metas):
            mask = torch.Tensor([1 for _ in range(self.num_classes)])
            if "ignored_labels" in meta and meta["ignored_labels"]:
                mask[meta["ignored_labels"]] = 0
            valid_label_mask.append(mask)
        valid_label_mask = torch.stack(valid_label_mask, dim=0)
        return valid_label_mask


class AnglularLinear(nn.Module):
    """Computes cos of angles between input vectors and weights vectors
    Args:
        in_features (int): Number of input features.
        out_features (int): Number of output cosine logits.
    """

    def __init__(self, in_features, out_features):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = nn.Parameter(torch.Tensor(self.out_features, self.in_features))
        self.weight.data.normal_().renorm_(2, 0, 1e-5).mul_(1e5)

    def forward(self, x):
        cos_theta = F.normalize(x.view(x.shape[0], -1), dim=1).mm(F.normalize(self.weight.t(), p=2, dim=0))
        return cos_theta.clamp(-1, 1)



"""NNCF wrapped mmcls models builder."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from functools import partial
from typing import Optional, Union

import torch
from mmcv.parallel import DataContainer
from mmcv.runner import CheckpointLoader
from mmcv.utils import Config, ConfigDict

# pylint: disable=no-name-in-module
from mmseg.utils import get_root_logger  # type: ignore

from otx.algorithms.common.adapters.mmcv.nncf.runners import NNCF_META_KEY
from otx.algorithms.common.adapters.mmcv.utils import (
    get_configs_by_pairs,
    remove_from_configs_by_type,
)
from otx.algorithms.common.adapters.nncf import is_accuracy_aware_training_set
from otx.algorithms.common.adapters.nncf.compression import NNCFMetaState
from otx.algorithms.segmentation.adapters.mmseg.utils import build_segmentor

logger = get_root_logger()


def build_nncf_segmentor(  # noqa: C901  # pylint: disable=too-many-locals,too-many-statements
    config: Config,
    train_cfg: Optional[Union[Config, ConfigDict]] = None,
    test_cfg: Optional[Union[Config, ConfigDict]] = None,
    checkpoint: Optional[str] = None,
    device: Union[str, torch.device] = "cpu",
    cfg_options: Optional[Union[Config, ConfigDict]] = None,
    distributed=False,
):
    """A function to build NNCF wrapped mmcls model."""

    from mmseg.apis import multi_gpu_test, single_gpu_test
    from mmseg.apis.inference import LoadImage
    from mmseg.datasets import build_dataloader as mmseg_build_dataloader
    from mmseg.datasets import build_dataset as mmseg_build_dataset
    from mmseg.datasets.pipelines import Compose

    from otx.algorithms.common.adapters.mmcv.nncf.utils import (
        get_fake_input,
        model_eval,
        wrap_nncf_model,
    )
    from otx.algorithms.common.adapters.mmcv.utils.builder import (
        build_dataloader,
        build_dataset,
    )

    if checkpoint is None:
        # load model in this function not in runner
        checkpoint = config.get("load_from")
    assert checkpoint is not None, "checkpoint is not given. NNCF model must be initialized with pretrained model"

    model = build_segmentor(
        config,
        train_cfg=train_cfg,
        test_cfg=test_cfg,
        cfg_options=cfg_options,
        from_scratch=True,
    )
    model = model.to(device)

    state_dict = CheckpointLoader.load_checkpoint(checkpoint, map_location=device)

    is_acc_aware = is_accuracy_aware_training_set(config.get("nncf_config"))

    init_dataloader = None
    model_eval_fn = None
    if "meta" in state_dict and NNCF_META_KEY in state_dict["meta"]:
        # NNCF ckpt
        nncf_meta_state = state_dict["meta"][NNCF_META_KEY]
        data_to_build_nncf = nncf_meta_state.data_to_build
        state_to_build_nncf = nncf_meta_state.state_to_build
    else:
        # pytorch ckpt
        state_to_build_nncf = state_dict
        if "state_dict" in state_dict:
            state_to_build_nncf = state_dict["state_dict"]

        init_dataloader = build_dataloader(
            build_dataset(
                config,
                subset="train",
                dataset_builder=mmseg_build_dataset,
            ),
            config,
            subset="train",
            dataloader_builder=mmseg_build_dataloader,
            distributed=distributed,
        )

        # This data and state dict will be used to build NNCF graph later
        # when loading NNCF model
        # because some models run their subcomponents based on intermediate outputs
        # resulting differently and partially traced NNCF graph
        data_to_build_nncf = next(iter(init_dataloader))["img"]
        if isinstance(data_to_build_nncf, DataContainer):
            data_to_build_nncf = data_to_build_nncf.data[0]
        data_to_build_nncf = data_to_build_nncf.cpu().numpy()
        if len(data_to_build_nncf.shape) == 4:
            data_to_build_nncf = data_to_build_nncf[0]
        if data_to_build_nncf.shape[0] == 3:
            data_to_build_nncf = data_to_build_nncf.transpose(1, 2, 0)

        val_dataloader = None
        if is_acc_aware:
            val_dataloader = build_dataloader(
                build_dataset(
                    config,
                    subset="val",
                    dataset_builder=mmseg_build_dataset,
                ),
                config,
                subset="val",
                dataloader_builder=mmseg_build_dataloader,
                distributed=distributed,
                # segmentor does not support various sized batch images
                samples_per_gpu=1,
            )

        model_eval_fn = partial(
            model_eval,
            config=config,
            val_dataloader=val_dataloader,
            evaluate_fn=multi_gpu_test if distributed else single_gpu_test,
            distributed=distributed,
        )
        state_dict = None

    test_pipeline = [LoadImage()] + config.data.test.pipeline[1:]
    test_pipeline = Compose(test_pipeline)
    get_fake_input_fn = partial(
        get_fake_input,
        preprocessor=test_pipeline,
        data=data_to_build_nncf,
    )

    compression_ctrl, model = wrap_nncf_model(
        config,
        model,
        model_eval_fn=model_eval_fn,
        get_fake_input_fn=get_fake_input_fn,
        dataloader_for_init=init_dataloader,
        is_accuracy_aware=is_acc_aware,
    )

    # update runner to save metadata
    config.runner.nncf_meta = NNCFMetaState(
        state_to_build=state_to_build_nncf,
        data_to_build=data_to_build_nncf,
    )

    # update custom hooks
    custom_hooks = config.get("custom_hooks", [])
    custom_hooks.append(ConfigDict({"type": "CancelTrainingHook"}))
    custom_hooks.append(
        ConfigDict(
            type="CompressionHook",
            compression_ctrl=compression_ctrl,
        )
    )
    # TODO: move this to OTX task when MPA is absorbed into OTX
    remove_from_configs_by_type(custom_hooks, "CancelInterfaceHook")
    remove_from_configs_by_type(custom_hooks, "TaskAdaptHook")
    remove_from_configs_by_type(custom_hooks, "LazyEarlyStoppingHook")
    remove_from_configs_by_type(custom_hooks, "EarlyStoppingHook")
    config.custom_hooks = custom_hooks

    for hook in get_configs_by_pairs(custom_hooks, dict(type="OTXProgressHook")):
        time_monitor = hook.get("time_monitor", None)
        if time_monitor and getattr(time_monitor, "on_initialization_end", None) is not None:
            time_monitor.on_initialization_end()

    return compression_ctrl, model



"""Tiling Module."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import copy
from itertools import product
from typing import Any, List, Tuple, Union

import numpy as np

from otx.api.utils.detection_utils import detection2array
from otx.api.utils.nms import multiclass_nms


class Tiler:
    """Tile Image into (non)overlapping Patches. Images are tiled in order to efficiently process large images.

    Args:
        tile_size: Tile dimension for each patch
        overlap: Overlap between adjacent tile
        max_number: max number of prediction per image
        segm: enable instance segmentation mask output
    """

    def __init__(
        self,
        tile_size: int,
        overlap: float,
        max_number: int,
        model: Any,
        segm: bool = False,
    ) -> None:
        self.tile_size = tile_size
        self.overlap = overlap
        self.stride = int(tile_size * (1 - overlap))
        self.max_number = max_number
        self.model = model
        self.segm = segm

    def tile(self, image: np.ndarray) -> List[List[int]]:
        """Tiles an input image to either overlapping, non-overlapping or random patches.

        Args:
            image: Input image to tile.

        Returns:
            Tiles coordinates
        """
        height, width = image.shape[:2]

        coords = [[0, 0, width, height]]
        for (loc_j, loc_i) in product(
            range(0, width - self.tile_size + 1, self.stride),
            range(0, height - self.tile_size + 1, self.stride),
        ):
            coords.append([loc_j, loc_i, loc_j + self.tile_size, loc_i + self.tile_size])
        return coords

    def predict(self, image: np.ndarray):
        """Predict by cropping full image to tiles.

        Args:
            image (np.ndarray): full size image

        Returns:
            detection: prediction results
            features: saliency map and feature vector
        """
        detections = np.empty((0, 6), dtype=np.float32)
        features = (None, None)
        masks: List[np.ndarray] = []
        for i, coord in enumerate(self.tile(image)):
            feats, output = self.predict_tile(image, coord, masks, i == 0)
            detections = np.append(detections, output, axis=0)
            # cache full image feature vector and saliency map at 0 index
            if i == 0:
                features = copy.deepcopy(feats)

        if np.prod(detections.shape):
            detections, keep = multiclass_nms(detections, max_num=self.max_number)
            if self.segm:
                masks = [masks[keep_idx] for keep_idx in keep]
                self.resize_masks(masks, detections, image.shape)
                detections = *Tiler.detection2tuple(detections), masks
        return detections, features

    def resize_masks(self, masks: List, dets: np.ndarray, shape: List[int]):
        """Resize Masks.

        Args:
            masks (List): list of raw np.ndarray masks
            dets (np.ndarray): detections including labels, scores, and boxes
            shape (List[int]): original full-res image shape
        """
        for i, (det, mask) in enumerate(zip(dets, masks)):
            masks[i] = self.model.segm_postprocess(det[2:], mask, *shape[:-1])

    def predict_tile(
        self,
        image: np.ndarray,
        coord: List[int],
        masks: List[np.ndarray],
        return_features=False,
    ):
        """Predict on single tile.

        Args:
            image (np.ndarray): full-res image
            coord (List): tile coordinates
            masks (List): list of raw np.ndarray masks
            return_features (bool, optional): return saliency map and feature vector if set to true. Defaults to False.

        Returns:
            features: saliency map and feature vector
            output: single tile prediction
        """
        features = (None, None)
        offset_x, offset_y, tile_dict, tile_meta = self.preprocess_tile(image, coord)
        raw_predictions = self.model.infer_sync(tile_dict)
        output = self.model.postprocess(raw_predictions, tile_meta)
        output = self.postprocess_tile(output, offset_x, offset_y, masks)
        if return_features:
            if "feature_vector" in raw_predictions or "saliency_map" in raw_predictions:
                features = (
                    raw_predictions["feature_vector"].reshape(-1),
                    raw_predictions["saliency_map"][0],
                )
        return features, output

    def postprocess_tile(
        self,
        output: Union[List, Tuple],
        offset_x: int,
        offset_y: int,
        masks: List,
    ):
        """Postprocess tile predictions.

        Args:
            output (Union[List, Tuple]): predictions
            offset_x (int): tile offset x value
            offset_y (int): tile offset y value
            masks (List): list of raw np.ndarray mask

        Returns:
            output: processed tile prediction
        """
        if self.segm:
            tile_scores, tile_labels, tile_boxes, tile_masks = output
            tile_boxes += np.tile([offset_x, offset_y], 2)
            out = np.concatenate(
                (
                    tile_labels[:, np.newaxis],
                    tile_scores[:, np.newaxis],
                    tile_boxes,
                ),
                -1,
            )
            masks.extend(tile_masks)
        else:
            assert isinstance(output, list)
            out = detection2array(output)
            out[:, 2:] += np.tile([offset_x, offset_y], 2)
        return out

    def preprocess_tile(self, image: np.ndarray, coord: List[int]):
        """Preprocess Tile by cropping.

        Args:
            image (np.ndarray): full-res image
            coord (List): tile coordinates

        Returns:
            _type_: _description_
        """
        x1, y1, x2, y2 = coord
        tile_dict, tile_meta = self.model.preprocess(image[y1:y2, x1:x2])
        if self.segm:
            tile_meta["resize_mask"] = False
        return x1, y1, tile_dict, tile_meta

    @staticmethod
    def detection2tuple(detections: np.ndarray):
        """_summary_.

        Args:
            detections (np.ndarray): _description_

        Returns:
            scores (np.ndarray): scores between 0-1
            labels (np.ndarray): label indices
            boxes (np.ndarray): boxes
        """
        labels = detections[:, 0]
        scores = detections[:, 1]
        boxes = detections[:, 2:]
        return scores, labels, boxes



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import mmcv
import numpy as np
from mmcv.parallel import DataContainer as DC
from mmseg.datasets import PIPELINES
from mmseg.datasets.pipelines.formatting import to_tensor


@PIPELINES.register_module(force=True)
class Normalize(object):
    """Normalize the image.

    Added key is "img_norm_cfg".

    Args:
        mean (sequence): Mean values of 3 channels.
        std (sequence): Std values of 3 channels.
        to_rgb (bool): Whether to convert the image from BGR to RGB,
            default is true.
    """

    def __init__(self, mean, std, to_rgb=True):
        self.mean = np.array(mean, dtype=np.float32)
        self.std = np.array(std, dtype=np.float32)
        self.to_rgb = to_rgb

    def __call__(self, results):
        """Call function to normalize images.

        Args:
            results (dict): Result dict from loading pipeline.

        Returns:
            dict: Normalized results, 'img_norm_cfg' key is added into
                result dict.
        """

        for target in ["img", "ul_w_img", "aux_img"]:
            if target in results:
                results[target] = mmcv.imnormalize(results[target], self.mean, self.std, self.to_rgb)
        results["img_norm_cfg"] = dict(mean=self.mean, std=self.std, to_rgb=self.to_rgb)

        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += f"(mean={self.mean}, std={self.std}, to_rgb=" f"{self.to_rgb})"
        return repr_str


@PIPELINES.register_module(force=True)
class DefaultFormatBundle(object):
    """Default formatting bundle.

    It simplifies the pipeline of formatting common fields, including "img"
    and "gt_semantic_seg". These fields are formatted as follows.

    - img: (1)transpose, (2)to tensor, (3)to DataContainer (stack=True)
    - gt_semantic_seg: (1)unsqueeze dim-0 (2)to tensor,
                       (3)to DataContainer (stack=True)
    """

    def __call__(self, results):
        """Call function to transform and format common fields in results.

        Args:
            results (dict): Result dict contains the data to convert.

        Returns:
            dict: The result dict contains the data that is formatted with
                default bundle.
        """
        for target in ["img", "ul_w_img", "aux_img"]:
            if target not in results:
                continue

            img = results[target]
            if len(img.shape) < 3:
                img = np.expand_dims(img, -1)

            if len(img.shape) == 3:
                img = np.ascontiguousarray(img.transpose(2, 0, 1)).astype(np.float32)
            elif len(img.shape) == 4:
                # for selfsl or supcon
                img = np.ascontiguousarray(img.transpose(0, 3, 1, 2)).astype(np.float32)
            else:
                raise ValueError(f"img.shape={img.shape} is not supported.")

            results[target] = DC(to_tensor(img), stack=True)

        for trg_name in ["gt_semantic_seg", "gt_class_borders", "pixel_weights"]:
            if trg_name not in results:
                continue

            out_type = np.float32 if trg_name == "pixel_weights" else np.int64
            results[trg_name] = DC(to_tensor(results[trg_name][None, ...].astype(out_type)), stack=True)

        return results

    def __repr__(self):
        return self.__class__.__name__


@PIPELINES.register_module()
class BranchImage(object):
    def __init__(self, key_map={}):
        self.key_map = key_map

    def __call__(self, results):
        for k1, k2 in self.key_map.items():
            if k1 in results:
                results[k2] = results[k1]
            if k1 in results["img_fields"]:
                results["img_fields"].append(k2)
        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        return repr_str



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import random
from typing import Union

import numpy as np
from PIL import Image, ImageEnhance, ImageOps
from PIL.Image import Resampling

import otx.mpa.modules.datasets.pipelines.transforms.cython_augments.pil_augment as pil_aug

PILImage = Image.Image
CvImage = np.ndarray
ImgTypes = Union[PILImage, CvImage]


class Augments:
    def _check_args_tf(kwargs):
        def _interpolation(kwargs):
            interpolation = kwargs.pop("resample", Resampling.BILINEAR)
            if isinstance(interpolation, (list, tuple)):
                return random.choice(interpolation)
            else:
                return interpolation

        kwargs["resample"] = _interpolation(kwargs)

    @staticmethod
    def autocontrast(img: PILImage, *args, **kwargs) -> PILImage:
        return ImageOps.autocontrast(img)

    @staticmethod
    def equalize(img: PILImage, *args, **kwargs) -> PILImage:
        return ImageOps.equalize(img)

    @staticmethod
    def solarize(img: PILImage, threshold: int, *args, **kwargs) -> PILImage:
        return ImageOps.solarize(img, threshold)

    @staticmethod
    def posterize(img: PILImage, bits_to_keep: int, *args, **kwargs) -> PILImage:
        if bits_to_keep >= 8:
            return img

        return ImageOps.posterize(img, bits_to_keep)

    @staticmethod
    def color(img: PILImage, factor: float, *args, **kwargs) -> PILImage:
        return ImageEnhance.Color(img).enhance(factor)

    @staticmethod
    def contrast(img: PILImage, factor: float, *args, **kwargs) -> PILImage:
        return ImageEnhance.Contrast(img).enhance(factor)

    @staticmethod
    def brightness(img: PILImage, factor: float, *args, **kwargs) -> PILImage:
        return ImageEnhance.Brightness(img).enhance(factor)

    @staticmethod
    def sharpness(img: PILImage, factor: float, *args, **kwargs) -> PILImage:
        return ImageEnhance.Sharpness(img).enhance(factor)

    @staticmethod
    def rotate(img: PILImage, degree: float, *args, **kwargs) -> PILImage:
        Augments._check_args_tf(kwargs)
        return img.rotate(degree, **kwargs)

    @staticmethod
    def shear_x(img: PILImage, factor: float, *args, **kwargs) -> PILImage:
        Augments._check_args_tf(kwargs)
        return img.transform(img.size, Image.AFFINE, (1, factor, 0, 0, 1, 0), **kwargs)

    @staticmethod
    def shear_y(img: PILImage, factor: float, *args, **kwargs) -> PILImage:
        Augments._check_args_tf(kwargs)
        return img.transform(img.size, Image.AFFINE, (1, 0, 0, factor, 1, 0), **kwargs)

    @staticmethod
    def translate_x_rel(img: PILImage, pct: float, *args, **kwargs) -> PILImage:
        Augments._check_args_tf(kwargs)
        pixels = pct * img.size[0]
        return img.transform(img.size, Image.AFFINE, (1, 0, pixels, 0, 1, 0), **kwargs)

    @staticmethod
    def translate_y_rel(img: PILImage, pct: float, *args, **kwargs) -> PILImage:
        Augments._check_args_tf(kwargs)
        pixels = pct * img.size[1]
        return img.transform(img.size, Image.AFFINE, (1, 0, 0, 0, 1, pixels), **kwargs)


class CythonAugments(Augments):
    def autocontrast(img: ImgTypes, *args, **kwargs) -> ImgTypes:
        if Image.isImageType(img):
            return pil_aug.autocontrast(img)

        raise NotImplementedError(f"Unknown type: {type(img)}")

    def equalize(img: ImgTypes, *args, **kwargs) -> ImgTypes:
        if Image.isImageType(img):
            return pil_aug.equalize(img)

        raise NotImplementedError(f"Unknown type: {type(img)}")

    def solarize(img: ImgTypes, threshold: int, *args, **kwargs) -> ImgTypes:
        if Image.isImageType(img):
            return pil_aug.solarize(img, threshold)

        raise NotImplementedError(f"Unknown type: {type(img)}")

    def posterize(img: ImgTypes, bits_to_keep: int, *args, **kwargs) -> ImgTypes:
        if Image.isImageType(img):
            if bits_to_keep >= 8:
                return img

            return pil_aug.posterize(img, bits_to_keep)

        raise NotImplementedError(f"Unknown type: {type(img)}")

    def color(img: ImgTypes, factor: float, *args, **kwargs) -> ImgTypes:
        if Image.isImageType(img):
            return pil_aug.color(img, factor)

        raise NotImplementedError(f"Unknown type: {type(img)}")

    def contrast(img: ImgTypes, factor: float, *args, **kwargs) -> ImgTypes:
        if Image.isImageType(img):
            return pil_aug.contrast(img, factor)

        raise NotImplementedError(f"Unknown type: {type(img)}")

    def brightness(img: ImgTypes, factor: float, *args, **kwargs) -> ImgTypes:
        if Image.isImageType(img):
            return pil_aug.brightness(img, factor)

        raise NotImplementedError(f"Unknown type: {type(img)}")

    def sharpness(img: ImgTypes, factor: float, *args, **kwargs) -> ImgTypes:
        if Image.isImageType(img):
            return pil_aug.sharpness(img, factor)

        raise NotImplementedError(f"Unknown type: {type(img)}")

    def rotate(img: ImgTypes, degree: float, *args, **kwargs) -> ImgTypes:
        Augments._check_args_tf(kwargs)

        if Image.isImageType(img):
            return pil_aug.rotate(img, degree)

        raise NotImplementedError(f"Unknown type: {type(img)}")

    def shear_x(img: ImgTypes, factor: float, *args, **kwargs) -> ImgTypes:
        Augments._check_args_tf(kwargs)

        if Image.isImageType(img):
            return pil_aug.shear_x(img, factor)

        raise NotImplementedError(f"Unknown type: {type(img)}")

    def shear_y(img: ImgTypes, factor: float, *args, **kwargs) -> ImgTypes:
        if Image.isImageType(img):
            return pil_aug.shear_y(img, factor)

        raise NotImplementedError(f"Unknown type: {type(img)}")

    def translate_x_rel(img: ImgTypes, pct: float, *args, **kwargs) -> ImgTypes:
        if Image.isImageType(img):
            return pil_aug.translate_x_rel(img, pct)

        raise NotImplementedError(f"Unknown type: {type(img)}")

    def translate_y_rel(img: ImgTypes, pct: float, *args, **kwargs) -> ImgTypes:
        if Image.isImageType(img):
            return pil_aug.translate_y_rel(img, pct)

        raise NotImplementedError(f"Unknown type: {type(img)}")

    def blend(src: ImgTypes, dst: CvImage, weight: float):
        assert isinstance(dst, CvImage), f"Type of dst should be numpy array, but type(dst)={type(dst)}."

        if Image.isImageType(src):
            return pil_aug.blend(src, dst, weight)

        raise NotImplementedError(f"Unknown type: {type(src)}")



"""Initialization of inference interfaces."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .inference import (
    AsyncOpenVINOTask,
    BaseInferencer,
    BaseOpenVINOInferencer,
    IInferencer,
)

__all__ = [
    "AsyncOpenVINOTask",
    "BaseInferencer",
    "BaseOpenVINOInferencer",
    "IInferencer",
]



"""Base MMDataset for Segmentation Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from abc import ABCMeta
from typing import Any, Dict, List, Optional, Sequence

import numpy as np
from mmseg.datasets.builder import DATASETS
from mmseg.datasets.custom import CustomDataset
from mmseg.datasets.pipelines import Compose

from otx.algorithms.common.utils.data import get_old_new_img_indices
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label import LabelEntity
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    check_input_parameters_type,
)
from otx.api.utils.segmentation_utils import mask_from_dataset_item


# pylint: disable=invalid-name, too-many-locals, too-many-instance-attributes, super-init-not-called
@check_input_parameters_type()
def get_annotation_mmseg_format(dataset_item: DatasetItemEntity, labels: List[LabelEntity]) -> dict:
    """Function to convert a OTX annotation to mmsegmentation format.

    This is used both in the OTXDataset class defined in this file
    as in the custom pipeline element 'LoadAnnotationFromOTXDataset'

    :param dataset_item: DatasetItem for which to get annotations
    :param labels: List of labels in the project
    :return dict: annotation information dict in mmseg format
    """

    gt_seg_map = mask_from_dataset_item(dataset_item, labels)
    gt_seg_map = gt_seg_map.squeeze(2).astype(np.uint8)

    ann_info = dict(gt_semantic_seg=gt_seg_map)

    return ann_info


@DATASETS.register_module()
class OTXSegDataset(CustomDataset, metaclass=ABCMeta):
    """Wrapper that allows using a OTX dataset to train mmsegmentation models.

    This wrapper is not based on the filesystem,
    but instead loads the items here directly from the OTX Dataset object.

    The wrapper overwrites some methods of the CustomDataset class: prepare_train_img, prepare_test_img and prepipeline
    Naming of certain attributes might seem a bit peculiar but this is due to the conventions set in CustomDataset. For
    instance, CustomDatasets expects the dataset items to be stored in the attribute data_infos, which is why it is
    named like that and not dataset_items.

    """

    class _DataInfoProxy:
        """This class is intended to be a wrapper to use it in CustomDataset-derived class as `self.data_infos`.

        Instead of using list `data_infos` as in CustomDataset, our implementation of dataset OTXDataset
        uses this proxy class with overriden __len__ and __getitem__; this proxy class
        forwards data access operations to otx_dataset and converts the dataset items to the view
        convenient for mmsegmentation.
        """

        def __init__(
            self,
            otx_dataset,
            labels=None,
            **kwargs,  # pylint: disable=unused-argument
        ):
            self.otx_dataset = otx_dataset
            self.labels = labels
            self.label_idx = {label.id: i for i, label in enumerate(labels)}

        def __len__(self):
            return len(self.otx_dataset)

        def __getitem__(self, index):
            """Prepare a dict 'data_info' that is expected by the mmseg pipeline to handle images and annotations.

            :return data_info: dictionary that contains the image and image metadata, as well as the labels of
            the objects in the image
            """
            dataset = self.otx_dataset
            item = dataset[index]
            ignored_labels = np.array([self.label_idx[lbs.id] + 1 for lbs in item.ignored_labels])

            data_info = dict(
                dataset_item=item,
                width=item.width,
                height=item.height,
                index=index,
                ann_info=dict(labels=self.labels),
                ignored_labels=ignored_labels,
            )

            return data_info

    @check_input_parameters_type({"otx_dataset": DatasetParamTypeCheck})
    def __init__(
        self,
        otx_dataset: DatasetEntity,
        pipeline: Sequence[dict],
        classes: Optional[List[str]] = None,
        test_mode: bool = False,
    ):
        self.otx_dataset = otx_dataset
        self.test_mode = test_mode

        self.ignore_index = 255
        self.reduce_zero_label = False
        self.label_map = None

        dataset_labels = self.otx_dataset.get_labels(include_empty=False)
        self.project_labels = self.filter_labels(dataset_labels, classes)
        self.CLASSES, self.PALETTE = self.get_classes_and_palette(classes, None)

        # Instead of using list data_infos as in CustomDataset, this implementation of dataset
        # uses a proxy class with overriden __len__ and __getitem__; this proxy class
        # forwards data access operations to otx_dataset.
        # Note that list `data_infos` cannot be used here, since OTX dataset class does not have interface to
        # get only annotation of a data item, so we would load the whole data item (including image)
        # even if we need only checking aspect ratio of the image; due to it
        # this implementation of dataset does not uses such tricks as skipping images with wrong aspect ratios or
        # small image size, since otherwise reading the whole dataset during initialization will be required.
        self.data_infos = OTXSegDataset._DataInfoProxy(self.otx_dataset, self.project_labels)

        self.pipeline = Compose(pipeline)

    @staticmethod
    @check_input_parameters_type()
    def filter_labels(all_labels: List[LabelEntity], label_names: List[str]):
        """Filtering Labels function."""
        filtered_labels = []
        for label_name in label_names:
            matches = [label for label in all_labels if label.name == label_name]
            if len(matches) == 0:
                continue

            assert len(matches) == 1

            filtered_labels.append(matches[0])

        return filtered_labels

    def __len__(self):
        """Total number of samples of data."""

        return len(self.data_infos)

    @check_input_parameters_type()
    def pre_pipeline(self, results: Dict[str, Any]):
        """Prepare results dict for pipeline."""

        results["seg_fields"] = []

    @check_input_parameters_type()
    def prepare_train_img(self, idx: int) -> dict:
        """Get training data and annotations after pipeline.

        Args:
            idx (int): Index of data.

        Returns:
            dict: Training data and annotation after pipeline with new keys introduced by pipeline.
        """

        item = self.data_infos[idx]

        self.pre_pipeline(item)
        out = self.pipeline(item)

        return out

    @check_input_parameters_type()
    def prepare_test_img(self, idx: int) -> dict:
        """Get testing data after pipeline.

        Args:
            idx (int): Index of data.

        Returns:
            dict: Testing data after pipeline with new keys introduced by pipeline.
        """

        item = self.data_infos[idx]

        self.pre_pipeline(item)
        out = self.pipeline(item)

        return out

    @check_input_parameters_type()
    def get_ann_info(self, idx: int):
        """This method is used for evaluation of predictions.

        The CustomDataset class implements a method
        CustomDataset.evaluate, which uses the class method get_ann_info to retrieve annotations.

        :param idx: index of the dataset item for which to get the annotations
        :return ann_info: dict that contains the coordinates of the bboxes and their corresponding labels
        """

        dataset_item = self.otx_dataset[idx]
        ann_info = get_annotation_mmseg_format(dataset_item, self.project_labels)

        return ann_info

    @check_input_parameters_type()
    def get_gt_seg_maps(self, efficient_test: bool = False):
        """Get ground truth segmentation maps for evaluation."""

        gt_seg_maps = []
        for item_id in range(len(self)):
            ann_info = self.get_ann_info(item_id)
            gt_seg_maps.append(ann_info["gt_semantic_seg"])
        if efficient_test:
            pass

        return gt_seg_maps


@DATASETS.register_module()
class MPASegDataset(OTXSegDataset, metaclass=ABCMeta):
    """Wrapper dataset that allows using a OTX dataset to train models."""

    def __init__(self, **kwargs):
        pipeline = []
        test_mode = kwargs.get("test_mode", False)
        if "dataset" in kwargs:
            dataset = kwargs["dataset"]
            otx_dataset = dataset.otx_dataset
            pipeline = dataset.pipeline
            classes = dataset.labels
            new_classes = dataset.new_classes
        else:
            otx_dataset = kwargs["otx_dataset"]
            pipeline = kwargs["pipeline"]
            classes = kwargs["labels"]
            new_classes = kwargs.get("new_classes", [])

        if test_mode is False:
            self.img_indices = get_old_new_img_indices(classes, new_classes, otx_dataset)

        if classes:
            classes = [c.name for c in classes]
            classes = ["background"] + classes
        else:
            classes = []
        super().__init__(otx_dataset=otx_dataset, pipeline=pipeline, classes=classes)
        if self.label_map is None:
            self.label_map = {}
            for i, c in enumerate(self.CLASSES):
                if c not in classes:
                    self.label_map[i] = -1
                else:
                    self.label_map[i] = classes.index(c)



_base_ = [
    "../_base_/default.py",
    "../_base_/data/data.py",
    "../_base_/logs/tensorboard_logger.py",
    "../_base_/optimizers/sgd.py",
    "../_base_/runners/epoch_runner_cancel.py",
    "../_base_/schedules/plateau.py",
]

# Disabling until evaluation bug resolved
# data = dict(
#     val=dict(samples_per_gpu=2),
#     test=dict(samples_per_gpu=2),
# )

optimizer = dict(
    lr=0.001,
    momentum=0.9,
    weight_decay=0.0001,
)

optimizer_config = dict(
    _delete_=True,
    type="OptimizerHook",
    #  type="SAMOptimizerHook",
    grad_clip=dict(max_norm=35, norm_type=2),
)

lr_config = dict(min_lr=1e-06)

evaluation = dict(interval=1, metric="bbox", classwise=True, save_best="bbox_mAP")

custom_hooks = [
    dict(
        type="LazyEarlyStoppingHook",
        start=3,
        patience=10,
        iteration_patience=0,
        metric="bbox_mAP",
        interval=1,
        priority=75,
    ),
]



"""Collection of helper functions for unit tests of otx.algorithms.detection."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import json
import os
from typing import Any, Dict, List

import numpy as np

from otx.algorithms.detection.utils import generate_label_schema
from otx.api.entities.annotation import AnnotationSceneEntity, AnnotationSceneKind
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.id import ID
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.model_template import TaskType, task_type_to_label_domain
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.utils.shape_factory import ShapeFactory
from tests.test_helpers import generate_random_annotated_image

DEFAULT_DET_MODEL_CONFIG_PATH = "otx/algorithms/detection/configs/detection/mobilenetv2_atss/model.py"
DEFAULT_ISEG_MODEL_CONFIG_PATH = (
    "otx/algorithms/detection/configs/instance_segmentation/efficientnetb2b_maskrcnn/model.py"
)

DEFAULT_DET_TEMPLATE_DIR = os.path.join("otx/algorithms/detection/configs/detection", "mobilenetv2_atss")
DEFAULT_ISEG_TEMPLATE_DIR = os.path.join(
    "otx/algorithms/detection/configs/instance_segmentation", "efficientnetb2b_maskrcnn"
)
DEFAULT_DET_RECIPE_CONFIG_PATH = "otx/recipes/stages/detection/incremental.py"
DEFAULT_ISEG_RECIPE_CONFIG_PATH = "otx/recipes/stages/instance-segmentation/incremental.py"


class MockImage(Image):
    """Mock class for Image entity."""

    @property
    def numpy(self) -> np.ndarray:
        """Returns empty numpy array"""

        return np.ndarray((256, 256))


class MockPipeline:
    """Mock class for data pipeline.

    It returns its inputs.
    """

    def __call__(self, results: Dict[str, Any]) -> Dict[str, Any]:
        return results


def init_environment(params, model_template, task_type=TaskType.DETECTION):
    classes = ("rectangle", "ellipse", "triangle")
    label_schema = generate_label_schema(classes, task_type_to_label_domain(task_type))
    environment = TaskEnvironment(
        model=None,
        hyper_parameters=params,
        label_schema=label_schema,
        model_template=model_template,
    )
    return environment


def generate_det_dataset(task_type, number_of_images=1):
    classes = ("rectangle", "ellipse", "triangle")
    label_schema = generate_label_schema(classes, task_type_to_label_domain(task_type))

    items = []
    for _ in range(number_of_images):
        image_numpy, annos = generate_random_annotated_image(
            image_width=640,
            image_height=480,
            labels=label_schema.get_labels(False),
        )
        # Convert shapes according to task
        for anno in annos:
            if task_type == TaskType.DETECTION:
                anno.shape = ShapeFactory.shape_as_rectangle(anno.shape)
            elif task_type == TaskType.INSTANCE_SEGMENTATION:
                anno.shape = ShapeFactory.shape_as_polygon(anno.shape)
        image = Image(data=image_numpy)
        annotation_scene = AnnotationSceneEntity(kind=AnnotationSceneKind.ANNOTATION, annotations=annos)
        items.append(DatasetItemEntity(media=image, annotation_scene=annotation_scene))
    dataset = DatasetEntity(items)
    return dataset, dataset.get_labels()


def generate_labels(length: int, domain: Domain) -> List[LabelEntity]:
    """Generate list of LabelEntity given length and domain."""

    output: List[LabelEntity] = []
    for i in range(length):
        output.append(LabelEntity(name=f"{i + 1}", domain=domain, id=ID(i + 1)))
    return output


def create_dummy_coco_json(json_name):
    image = {
        "id": 0,
        "width": 640,
        "height": 640,
        "file_name": "fake_name.jpg",
    }

    annotation_1 = {
        "id": 1,
        "image_id": 0,
        "category_id": 0,
        "area": 400,
        "bbox": [50, 60, 20, 20],
        "segmentation": [[165.16, 2.58, 344.95, 41.29, 27.5, 363.0, 9.46, 147.1]],
        "iscrowd": 0,
    }

    annotation_2 = {
        "id": 2,
        "image_id": 0,
        "category_id": 0,
        "area": 900,
        "bbox": [100, 120, 30, 30],
        "segmentation": [[165.16, 2.58, 344.95, 41.29, 27.5, 363.0, 9.46, 147.1]],
        "iscrowd": 0,
    }

    categories = [
        {
            "id": 0,
            "name": "car",
            "supercategory": "car",
        }
    ]

    fake_json = {
        "images": [image],
        "annotations": [annotation_1, annotation_2],
        "categories": categories,
    }
    with open(json_name, "w") as f:
        json.dump(fake_json, f)



"""This module implements time related utility functions."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


import datetime
import functools
import time
from typing import Optional


def now() -> datetime.datetime:
    """Return the current UTC creation_date and time up to a millisecond accuracy.

    This function is preferable over the Python datetime.datetime.now() function
    because it uses the same accuracy (milliseconds) as MongoDB rather than microsecond accuracy.

    Returns:
        Date and time up to a millisecond precision.
    """
    date = datetime.datetime.now(datetime.timezone.utc)
    return date.replace(microsecond=(date.microsecond // 1000) * 1000)


# Debug tools
def timeit(func):
    """This function can be used as a decorator as @timeit.

    It will print out how long the function took to execute.

    Args:
        func: The decorated function

    Returns:
        The wrapped function
    """

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        start_time = time.time()
        res = func(*args, **kwargs)
        elapsed_time = time.time() - start_time
        print(f"function [{func.__name__}] finished in {float(elapsed_time * 1000)} ms")
        return res

    return new_func


class TimeEstimator:
    """The time estimator.

    Estimate the remaining time given the progress, and the progress changes. The estimator starts estimation at a
    starting progress that is not necessarily 0. This choice is motivated by the fact that the first percent of
    progress often takes a much longer time than the following percents.

    Args:
        smoothing_factor (float): Smoothing factor for the exponentially
            weighted moving average. There's a great explanation at
            https://www.wallstreetmojo.com/ewma/
        inflation_factor (float): The factor by which the initial total time
            estimation is inflated to ensure decreasing
        update_window (float): Last update happened at progress1, next update
            will happen at (progress1 + update window)
        starting_progress (float): The progress at which the time_remaining
            estimation starts time estimation
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        smoothing_factor: float = 0.02,
        inflation_factor: float = 1.1,
        update_window: float = 1.0,
        starting_progress: float = 1.0,
    ):
        self.estimated_total_time: Optional[float] = None
        self.estimated_end_time: Optional[float] = None
        self.first_update_progress = starting_progress
        self.first_update_time: Optional[float] = None
        self.last_update_progress: Optional[float] = None
        self.estimated_remaining_time: Optional[float] = None
        self.starting_progress = starting_progress
        self.smoothing_factor = smoothing_factor
        self.inflation_factor = inflation_factor
        self.update_window = update_window

    def time_remaining_from_progress(self, progress: float) -> float:
        """Updates the current progress, and returns the estimated remaining time in seconds (float).

        Args:
            progress (float): The new progress (floating point percentage, 0.0 - 100.0)

        Returns:
            The expected remaining time in seconds (float)
        """
        estimation = -1.0
        if progress is not None and progress > 0:
            self.update(progress)
            estimation = self.get_time_remaining()
        return estimation

    def get_time_remaining(self):
        """If the new estimation is higher than the previous one by up to 2 seconds, return old estimation.

        Returns:
            Estimated remaining time in seconds (float)
        """
        new_estimation = self.estimated_end_time - time.time() if self.estimated_end_time is not None else -1.0
        if self.estimated_remaining_time is None or not 0.0 < new_estimation - self.estimated_remaining_time < 2.0:
            self.estimated_remaining_time = new_estimation
        return self.estimated_remaining_time

    def update(self, progress: float):
        """Update the estimator with a new progress (floating point percentage, between 0.0 - 100.0).

        Args:
            progress (float): Progress of the process

        Returns:
            None
        """
        if progress >= self.first_update_progress and self.last_update_progress is None:
            self.first_update_progress = progress
            self.last_update_progress = progress
            self.first_update_time = time.time()

        if self.last_update_progress is not None and progress - self.last_update_progress >= self.update_window:

            if self.first_update_time is None or self.first_update_progress is None:
                raise AssertionError(
                    "first_update_time and first_update_progress both can not be None when calling "
                    "TimeEstimator.update()."
                )

            self.last_update_progress = progress
            # normalized progress since starting point for estimation
            normalized_progress = (progress - self.first_update_progress) / (100 - self.first_update_progress)
            time_elapsed = time.time() - self.first_update_time
            estimated_total = time_elapsed / normalized_progress
            if self.estimated_total_time is None:
                # inflating the initial estimation to ensure decreasing time remaining
                self.estimated_total_time = estimated_total * self.inflation_factor
            else:
                self.estimated_total_time = (
                    self.smoothing_factor * self.estimated_total_time + (1 - self.smoothing_factor) * estimated_total
                )
            self.estimated_end_time = self.first_update_time + self.estimated_total_time



from copy import deepcopy

import pytest
import torch

from otx.mpa.modules.models.classifiers.sam_classifier import (
    ImageClassifier,
    SAMImageClassifier,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class OTXMobileNetV3:
    pass


class OTXEfficientNet:
    pass


class OTXEfficientNetV2:
    pass


class MockModule:
    def __init__(self, name):
        if name == "mobilenet":
            self.backbone = OTXMobileNetV3()
            self._state_dict = {
                "classifier.4.weight": torch.rand(1, 1),
                "classifier.4.bias": torch.rand(1),
                "act.weight": torch.rand(1),
                "someweights": torch.rand(2),
            }
        elif name == "effnetv2":
            self.backbone = OTXEfficientNetV2()
            self._state_dict = {
                "model.classifier.weight": torch.rand(1, 1),
                "model.weight": torch.rand(1),
            }
        elif name == "effnet":
            self.backbone = OTXEfficientNet()
            self._state_dict = {
                "features.weight": torch.rand(1, 1),
                "features.active.weight": torch.rand(1, 1),
                "output.weight": torch.rand(1),
                "output.asl.weight": torch.rand(1),
            }
        self.multilabel = False
        self.hierarchical = False
        self.is_export = False

    def state_dict(self):
        return self._state_dict


class TestSAMImageClassifier:
    @pytest.fixture(autouse=True)
    def setup(self, mocker) -> None:
        mocker.patch.object(ImageClassifier, "__init__", return_value=None)
        SAMImageClassifier._register_state_dict_hook = mocker.MagicMock()
        SAMImageClassifier._register_load_state_dict_pre_hook = mocker.MagicMock()
        self.classifier = SAMImageClassifier()

    @e2e_pytest_unit
    def test_forward_train(self, mocker):
        img = torch.rand(1, 3, 224, 224)
        gt_labels = torch.rand(1, 1)
        self.classifier.extract_feat = mocker.MagicMock()
        self.classifier.head = mocker.MagicMock()
        self.classifier.augments = None
        self.classifier.multilabel = False
        self.classifier.hierarchical = False
        losses = self.classifier.forward_train(img, gt_labels)

        assert losses is not None

    @pytest.mark.parametrize("name", ["mobilenet", "effnetv2", "effnet"])
    @e2e_pytest_unit
    def test_load_state_dict_pre_hook(self, name):
        self.module = MockModule(name)
        state_dict = self.module.state_dict()
        self.classifier.load_state_dict_pre_hook(self.module, state_dict, prefix="")

        for key in state_dict:
            if name == "mobilenet":
                if "classifier" in key:
                    assert ".3." in key
                if "someweights" in key:
                    assert "backbone." in key
                if "act" in key:
                    assert "head." in key
            elif name == "effnetv2":
                assert "classifier" not in key
                if "model" in key:
                    assert "backbone" in key
            else:
                if "features" in key and "active" not in key:
                    assert "backbone" in key
                elif "active" in key:
                    assert key == "features.active.weight"
                else:
                    assert "head" in key or "fc" in key

    @pytest.mark.parametrize("name", ["mobilenet", "effnetv2", "effnet"])
    @e2e_pytest_unit
    def test_state_dict_hook(self, name):
        self.module = MockModule(name)
        state_dict = self.module.state_dict()
        state_dict_copy = deepcopy(state_dict)
        self.classifier.load_state_dict_pre_hook(self.module, state_dict, prefix="")
        # backward state dict
        self.classifier.state_dict_hook(self.module, state_dict, prefix="")

        assert state_dict.keys() == state_dict_copy.keys()

    @pytest.mark.parametrize("name", ["mobilenet", "effnetv2", "effnet"])
    def test_load_state_dict_mixing_hook(self, name):
        self.module = MockModule(name)
        state_dict = self.module.state_dict()
        chkpt_dict = deepcopy(state_dict)
        model_classes = [0, 1, 2]
        chkpt_classes = [0, 1]
        self.classifier.load_state_dict_mixing_hook(self.module, model_classes, chkpt_classes, chkpt_dict, prefix="")

        assert chkpt_dict.keys() == state_dict.keys()



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
import torch.nn as nn
import torch.nn.functional as F
from mmcls.models.builder import HEADS
from mmcls.models.heads.cls_head import ClsHead
from mmcv.cnn import build_activation_layer, constant_init, normal_init


@HEADS.register_module()
class NonLinearClsHead(ClsHead):
    """Nonlinear classifier head.

    Args:
        num_classes (int): Number of categories excluding the background
            category.
        in_channels (int): Number of channels in the input feature map.
        hid_channels (int): Number of channels of hidden layer.
        act_cfg (dict): Config of activation layer.
        loss (dict): Config of classification loss.
        topk (int | tuple): Top-k accuracy.
    """  # noqa: W605

    def __init__(
        self,
        num_classes,
        in_channels,
        hid_channels=1280,
        act_cfg=dict(type="ReLU"),
        loss=dict(type="CrossEntropyLoss", loss_weight=1.0),
        topk=(1,),
        dropout=False,
        *args,
        **kwargs,
    ):
        topk = (1,) if num_classes < 5 else (1, 5)
        super(NonLinearClsHead, self).__init__(loss=loss, topk=topk, *args, **kwargs)
        self.in_channels = in_channels
        self.hid_channels = hid_channels
        self.num_classes = num_classes
        self.act = build_activation_layer(act_cfg)
        self.dropout = dropout

        if self.num_classes <= 0:
            raise ValueError(f"num_classes={num_classes} must be a positive integer")

        self._init_layers()

    def _init_layers(self):
        if self.dropout:
            self.classifier = nn.Sequential(
                nn.Linear(self.in_channels, self.hid_channels),
                nn.BatchNorm1d(self.hid_channels),
                self.act,
                nn.Dropout(p=0.2),
                nn.Linear(self.hid_channels, self.num_classes),
            )
        else:
            self.classifier = nn.Sequential(
                nn.Linear(self.in_channels, self.hid_channels),
                nn.BatchNorm1d(self.hid_channels),
                self.act,
                nn.Linear(self.hid_channels, self.num_classes),
            )

    def init_weights(self):
        for m in self.classifier:
            if isinstance(m, nn.Linear):
                normal_init(m, mean=0, std=0.01, bias=0)
            elif isinstance(m, nn.BatchNorm1d):
                constant_init(m, 1)

    def simple_test(self, img):
        """Test without augmentation."""
        cls_score = self.classifier(img)
        if isinstance(cls_score, list):
            cls_score = sum(cls_score) / float(len(cls_score))
        if torch.onnx.is_in_onnx_export():
            return cls_score
        pred = F.softmax(cls_score, dim=1) if cls_score is not None else None
        pred = list(pred.detach().cpu().numpy())
        return pred

    def forward_train(self, x, gt_label):
        cls_score = self.classifier(x)
        losses = self.loss(cls_score, gt_label)
        return losses



"""Segmentation models static deploy config."""

ir_config = dict(
    type="onnx",
    export_params=True,
    keep_initializers_as_inputs=False,
    opset_version=11,
    save_file="end2end.onnx",
    input_names=["input"],
    output_names=["output"],
    input_shape=None,
    # TODO
    # optimizing onnx graph mess up NNCF graph at some point
    # where we need to look into
    optimize=False,
)

codebase_config = dict(
    type="mmseg",
    task="Segmentation",
)

backend_config = dict(
    type="openvino",
    mo_options=None,
)

input_data = dict(
    shape=(128, 128, 3),
    file_path=None,
)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmcv.runner import HOOKS, Hook

from otx.mpa.utils.logger import get_logger

logger = get_logger()


@HOOKS.register_module()
class LoggerReplaceHook(Hook):
    """replace logger in the runner to the MPA logger.
    DO NOT INCLUDE this hook to the recipe directly.
    mpa will add this hook to all recipe internally.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def before_run(self, runner):
        runner.logger = logger
        logger.info("logger in the runner is replaced to the MPA logger")



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest
import torch

from otx.mpa.modules.models.heads.custom_cls_head import (
    CustomLinearClsHead,
    CustomNonLinearClsHead,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestLinearClsHead:
    @pytest.fixture(autouse=True)
    def head_type(self) -> None:
        return CustomLinearClsHead

    @pytest.fixture(autouse=True)
    def setup(self, head_type) -> None:
        self.num_classes = 2
        self.head_dim = 5
        self.default_head = head_type(self.num_classes, self.head_dim)
        self.default_input = torch.ones((2, self.head_dim))
        self.default_gt = torch.zeros(2).long()

    @e2e_pytest_unit
    def test_forward(self) -> None:
        result = self.default_head.forward_train(self.default_input, self.default_gt)
        assert "loss" in result
        assert result["loss"] >= 0

    @e2e_pytest_unit
    def test_forward_accuracy(self, head_type) -> None:
        head = head_type(self.num_classes, self.head_dim, cal_acc=True)
        result = head.forward_train(self.default_input, self.default_gt)
        assert "loss" in result
        assert "accuracy" in result
        assert "top-1" in result["accuracy"]
        assert result["accuracy"]["top-1"] >= 0

    @e2e_pytest_unit
    def test_simple_test(self) -> None:
        result = self.default_head.simple_test(self.default_input)
        assert result[0].shape[0] == self.num_classes


class TestNonLinearClsHead(TestLinearClsHead):
    @pytest.fixture(autouse=True)
    def head_type(self) -> None:
        return CustomNonLinearClsHead

    @e2e_pytest_unit
    def test_dropout(self, head_type) -> None:
        head = head_type(self.num_classes, self.head_dim, dropout=True)
        head.init_weights()
        assert len(head.classifier) > len(self.default_head.classifier)



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

"""
This file contains functions that may be used in the conftest file of algo
backend and in the standard pytest hooks:
* pytest_addoption
* pytest_generate_tests
to add to pytest functionality required for algo backends reallife training tests.
"""

try:
    import e2e.fixtures
    from e2e import config  # noqa
    from e2e.conftest_utils import *  # noqa
    from e2e.conftest_utils import pytest_addoption as _e2e_pytest_addoption  # noqa
    from e2e.utils import get_plugins_from_packages

    _pytest_plugins_from_e2e = get_plugins_from_packages([e2e])
except ImportError:
    _e2e_pytest_addoption = None
    _pytest_plugins_from_e2e = []


def get_pytest_plugins_from_otx():
    """
    The function generates pytest_plugins variable that should be used
    in an algo backend' conftest.py file.
    """
    import tests.test_suite.fixtures  # noqa

    pytest_plugins_from_otx_api = ["tests.test_suite.fixtures"]
    pytest_plugins = list(_pytest_plugins_from_e2e) + pytest_plugins_from_otx_api
    return pytest_plugins


def otx_pytest_addoption_insertion(parser):
    """
    The function should be called in the standard pytest hook pytest_addoption
    to add the options required for reallife training tests.
    """
    if _e2e_pytest_addoption:
        _e2e_pytest_addoption(parser)

    parser.addoption(
        "--dataset-definitions",
        action="store",
        default=None,
        help="Path to the dataset_definitions.yml file for tests that require datasets.",
    )
    parser.addoption(
        "--test-usecase",
        action="store",
        default=None,
        help="Optional. If the parameter is set, it filters test_otx_training tests by usecase field.",
    )
    parser.addoption(
        "--expected-metrics-file",
        action="store",
        default=None,
        help="Optional. If the parameter is set, it points the YAML file with expected test metrics.",
    )
    parser.addoption(
        "--force-log-level",
        action="store",
        default=None,
        help="Optional. If the parameter is set, the logger in each test is forced to this level.",
    )
    parser.addoption(
        "--force-log-level-recursive",
        action="store",
        default=None,
        help="Optional. If the parameter is set, the logger in each test and its parents " "are forced to this level.",
    )

    # TODO(lbeynens): remove it after update CI
    parser.addoption(
        "--template-paths",
        action="store",
        default=None,
        help="Obsolete parameter. Should be removed when CI is changed.",
    )


def otx_pytest_generate_tests_insertion(metafunc):
    """
    The function should be called in the standard pytest hook pytest_generate_tests
    in algo backend's conftest.py file to generate parameters of reallife training tests.
    """
    from .logging import get_logger
    from .training_tests_helper import OTXTrainingTestInterface

    logger = get_logger()
    if metafunc.cls is None:
        return False
    if not issubclass(metafunc.cls, OTXTrainingTestInterface):
        return False

    logger.debug(f"otx_pytest_generate_tests_insertion: begin handling {metafunc.cls}")

    # It allows to filter by usecase
    usecase = metafunc.config.getoption("--test-usecase")

    argnames, argvalues, ids = metafunc.cls.get_list_of_tests(usecase)

    assert isinstance(argnames, (list, tuple))
    assert "test_parameters" in argnames
    assert isinstance(argvalues, list)
    assert isinstance(ids, list)
    assert len(argvalues) == len(ids)
    assert all(isinstance(v, str) for v in ids)

    metafunc.parametrize(argnames, argvalues, ids=ids, scope="class")
    logger.debug(f"otx_pytest_generate_tests_insertion: end handling {metafunc.cls}")
    return True


def otx_conftest_insertion(*, default_repository_name=""):
    """
    The function should be called in an algo backend's conftest.py file
    to set default repository name in e2e- test system.
    """
    try:
        import os

        from e2e import config as config_e2e

        config_e2e.repository_name = os.environ.get("TT_REPOSITORY_NAME", default_repository_name)
    except ImportError:
        pass



"""AccuracyAwareRunner for NNCF task."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import time
from dataclasses import asdict

from mmcv.runner import RUNNERS
from mmcv.runner.hooks.evaluation import EvalHook
from mmcv.runner.hooks.lr_updater import LrUpdaterHook
from mmcv.runner.utils import get_host_info

from otx.algorithms.common.adapters.mmcv.nncf.hooks import CompressionHook
from otx.algorithms.common.adapters.mmcv.runner import EpochRunnerWithCancel
from otx.algorithms.common.adapters.nncf import (
    AccuracyAwareLrUpdater,
    check_nncf_is_enabled,
)
from otx.algorithms.common.adapters.nncf.compression import NNCFMetaState

NNCF_META_KEY = "nncf_meta"


# TODO: refactoring
@RUNNERS.register_module()
class AccuracyAwareRunner(EpochRunnerWithCancel):  # pylint: disable=too-many-instance-attributes
    """AccuracyAwareRunner for NNCF task.

    An mmcv training runner to be used with NNCF-based accuracy-aware training.
    Inherited from the standard EpochBasedRunner with the overridden "run" method.
    This runner does not use the "workflow" and "max_epochs" parameters that are
    used by the EpochBasedRunner since the training is controlled by NNCF's
    AdaptiveCompressionTrainingLoop that does the scheduling of the compression-aware
    training loop using the parameters specified in the "accuracy_aware_training".
    """

    def __init__(self, *args, nncf_config, nncf_meta=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.nncf_config = nncf_config

        if nncf_meta is None:
            nncf_meta = NNCFMetaState()
        self.nncf_meta = nncf_meta

        self.compression_ctrl = None
        self._target_metric_name = nncf_config["target_metric_name"]
        self._train_data_loader = None
        self._eval_hook = None

    def run(self, data_loaders, *args, **kwargs):  # pylint: disable=unused-argument
        """run."""
        check_nncf_is_enabled()

        from nncf.common.accuracy_aware_training import (
            create_accuracy_aware_training_loop,
        )

        assert isinstance(data_loaders, list)

        lr_update_hook = []
        eval_hook = []
        found_compression_hook = False
        for hook in self.hooks:
            if isinstance(hook, LrUpdaterHook):
                lr_update_hook.append(hook)
            if isinstance(hook, CompressionHook):
                found_compression_hook = True
            if isinstance(hook, EvalHook):
                eval_hook.append(hook)
        assert found_compression_hook, f"{CompressionHook} must be registered to {self}."
        assert len(lr_update_hook) <= 1, (
            f"More than 1 lr update hooks ({len(lr_update_hook)} " f"are registered to {self}"
        )
        assert len(eval_hook) == 1, f"{EvalHook} must be registered to {self}"
        self._eval_hook = eval_hook[0]
        assert self._eval_hook.save_best == self.nncf_config.target_metric_name, (
            "'target_metric_name' from nncf_config is not identical to 'save_best' in 'EvalHook'. "
            f"({self._eval_hook.save_best} != {self.nncf_config.target_metric_name})"
        )

        work_dir = self.work_dir if self.work_dir is not None else "NONE"
        self.logger.info("Start running, host: %s, work_dir: %s", get_host_info(), work_dir)
        self.logger.warning(
            "Note that the workflow and max_epochs parameters are not used in NNCF-based accuracy-aware training"
        )

        # taking only the first data loader for NNCF training
        self._train_data_loader = data_loaders[0]
        # Maximum possible number of iterations, needs for progress tracking
        params = self.nncf_config["accuracy_aware_training"]["params"]
        self._max_epochs = params["maximal_total_epochs"]
        self._max_iters = self._max_epochs * len(self._train_data_loader)

        self.logger.info("Start running, host: %s, work_dir: %s", get_host_info(), work_dir)
        self.logger.info("Hooks will be executed in the following order:\n%s", self.get_hook_info())
        self.call_hook("before_run")

        def configure_optimizers_fn():
            return self.optimizer, None

        if len(lr_update_hook) == 1:
            lr_update_hook = lr_update_hook[0]

            def configure_optimizers_fn():  # noqa: F811  # pylint: disable=function-redefined
                return self.optimizer, AccuracyAwareLrUpdater(lr_update_hook)

        # pylint: disable-next=unused-argument
        def dump_checkpoint_fn(model, compression_ctrl, nncf_runner, save_dir):
            # pylint: disable-next=protected-access
            self._eval_hook._save_ckpt(self, nncf_runner.best_val_metric_value)
            return self._eval_hook.best_ckpt_path

        acc_aware_training_loop = create_accuracy_aware_training_loop(
            self.nncf_config, self.compression_ctrl, verbose=False
        )

        model = acc_aware_training_loop.run(
            self.model,
            train_epoch_fn=self.train_fn,
            validate_fn=self.validation_fn,
            configure_optimizers_fn=configure_optimizers_fn,
            dump_checkpoint_fn=dump_checkpoint_fn,
            log_dir=self.work_dir,
        )

        time.sleep(1)  # wait for some hooks like loggers to finish
        self.call_hook("after_run")
        return model

    def train_fn(self, *args, **kwargs):  # pylint: disable=unused-argument
        """train_fn.

        Train the model for a single epoch.
        This method is used in NNCF-based accuracy-aware training.
        """
        self.train(self._train_data_loader)

    def validation_fn(self, *args, **kwargs):  # pylint: disable=unused-argument
        """validation_fn.

        Return the target metric value on the validation dataset.
        This method is used in NNCF-based accuracy-aware training.
        """

        # make sure evaluation hook is in a 'should_evaluate' state
        interval_bak = self._eval_hook.interval
        self._eval_hook.interval = 1
        self._eval_hook._do_evaluate(self)  # pylint: disable=protected-access
        self._eval_hook.interval = interval_bak
        # Get metric from runner's attributes that set in EvalHook.evaluate() function
        all_metrics = getattr(self, "all_metrics", {})
        metric = all_metrics.get(self._target_metric_name, None)
        if metric is None:
            raise RuntimeError(f"Could not find the {self._target_metric_name} key")
        return metric

    def save_checkpoint(self, *args, **kwargs) -> None:
        """Save checkpoint with NNCF meta state."""

        compression_state = self.compression_ctrl.get_compression_state()
        for algo_state in compression_state.get("ctrl_state", {}).values():
            if not algo_state.get("scheduler_state"):
                algo_state["scheduler_state"] = {"current_step": 0, "current_epoch": 0}

        nncf_meta = NNCFMetaState(
            **{**asdict(self.nncf_meta), "compression_ctrl": compression_state},
        )

        meta = kwargs.pop("meta", {})
        meta[NNCF_META_KEY] = nncf_meta
        meta["nncf_enable_compression"] = True
        super().save_checkpoint(*args, **kwargs, meta=meta)



"""This module implements the LabelSchema entity."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import copy
import logging
import re
from enum import Enum
from typing import Dict, List, Optional, Sequence, Union

import numpy as np
from bson import ObjectId

from otx.api.entities.graph import Graph, MultiDiGraph
from otx.api.entities.id import ID
from otx.api.entities.label import LabelEntity
from otx.api.entities.scored_label import ScoredLabel

logger = logging.getLogger(__name__)


def natural_sort_label_id(target: Union[ID, LabelEntity, ScoredLabel]) -> List:
    """Generates a natural sort key for a LabelEntity object based on its ID.

    Args:
    target (Union[ID, LabelEntity]): The ID or LabelEntity or ScoredLabel object to be sorted.

    Returns:
    List[int]: A list of integers representing the numeric substrings in the ID
    in the order they appear.

    Example:
    origin_sorted_labels = sorted(labels, key=lambda x: x.id_)
    natural_sorted_labels = sorted(labels, key=lambda x: x.natural_sort_label_id)

    print(origin_sorted_labels) # Output: [LabelEntity(0), LabelEntity(1), LabelEntity(10), ... LabelEntity(2)]
    print(natural_sorted_labels) # Output: [LabelEntity(0), LabelEntity(1), LabelEntity(2), ... LabelEntity(10)]
    """

    if isinstance(target, (LabelEntity, ScoredLabel)):
        target = target.id_
    if isinstance(target, int):
        return [target]
    return [int(t) if t.isdigit() else t for t in re.split(r"(\d+)", target)]


class LabelGroupExistsException(ValueError):
    """Exception thrown if the LabelGroup already exists."""


class LabelGroupDoesNotExistException(ValueError):
    """Exception thrown if the LabelGroup does not exist."""


class LabelGroupType(Enum):
    """Enum to indicate the LabelGroupType."""

    EXCLUSIVE = 1
    EMPTY_LABEL = 2


class LabelGroup:
    """A label group which has exclusive (multiclass) or contains the empty label.

    Non-exclusive (multilabel) relationships are represented by multiple (exclusive)
    label groups.

    The labels have to be from one task.

    Args:
        name (str): Descriptive name of the label group
        labels (Sequence[LabelEntity]): Labels that form the group
        group_type (LabelGroupType): EXCLUSIVE or EMPTY_LABEL
        id (ID): ID of the LabelGroup. If no ID is provided, a new ObjectId()
            will be assigned
    """

    # pylint: disable=redefined-builtin
    def __init__(
        self,
        name: str,
        labels: Sequence[LabelEntity],
        group_type: LabelGroupType = LabelGroupType.EXCLUSIVE,
        id: ID = None,
    ):
        self.id_ = ID(ObjectId()) if id is None else id

        self.labels = sorted(labels, key=natural_sort_label_id)
        self.name = name
        self.group_type = group_type

    @property
    def id(self) -> ID:
        """DEPRECATED."""
        return self.id_

    @id.setter
    def id(self, value: ID):
        """DEPRECATED."""
        self.id_ = value

    @property
    def minimum_label_id(self) -> ID:
        """Returns the minimum (oldest) label ID, which is the first label in self.labels since this list is sorted."""
        return self.labels[0].id_

    def remove_label(self, label: LabelEntity) -> None:
        """Remove label from label group if it exists in the group.

        Args:
            label (LabelEntity): label to remove
        """
        if label in self.labels:
            self.labels.remove(label)

    def is_single_label(self) -> bool:
        """Returns True if the label group only contains one label.

        Returns:
            bool: True if the label group only contains one label.
        """
        return len(self.labels) == 1

    def __eq__(self, other: object):
        """Returns True if the LabelGroup is equal to the other object."""
        if not isinstance(other, LabelGroup):
            return False
        return self.id_ == other.id_ and (set(self.labels) == set(other.labels) and self.group_type == other.group_type)

    def __repr__(self) -> str:
        """Returns the string representation of the LabelGroup."""
        return f"LabelGroup(id={self.id_}, name={self.name}, group_type={self.group_type}," f" labels={self.labels})"


class LabelGraph(Graph):
    """Represents connectivity between labels as a graph. For example exclusivity or hierarchy.

    Args:
        directed (bool): whether the relationships are directed or undirected
            (symmetrical)
    """

    def __init__(self, directed: bool):
        super().__init__(directed)

    def add_edges(self, edges):
        """Add edges between Labels."""
        self._graph.add_edges_from(edges)

    @property
    def num_labels(self):
        """Returns the number of nodes in the graph."""
        return self.num_nodes()

    @property
    def type(self):
        """Returns the type of the LabelGraph."""
        return "graph"

    def subgraph(self, labels: Sequence[LabelEntity]) -> "LabelGraph":
        """Return the subgraph containing the given labels."""
        new_graph = LabelGraph(self.directed)
        new_graph.set_graph(self.get_graph().subgraph(labels).copy())
        return new_graph

    def __eq__(self, other) -> bool:
        """Returns True if the LabelGraph is equal to the other LabelGraph."""
        if isinstance(other, LabelGraph):
            return super().__eq__(other)
        return False


class LabelTree(MultiDiGraph):
    """Represents a hierarchy of labels in the form a tree.

    The tree is represented by a directed graph
    """

    def __init__(self) -> None:
        super().__init__()

        self.__topological_order_cache: Optional[List[LabelEntity]] = None

    def add_edge(self, node1, node2, edge_value=None):
        """Add edge between two nodes in the tree.

        Args:
            node1: first node
            node2: second node
            edge_value: The value of the new edge. Defaults to None.
        """
        super().add_edge(node1, node2, edge_value)
        self.clear_topological_cache()

    def add_node(self, node):
        """Add node to the tree."""
        super().add_node(node)
        self.clear_topological_cache()

    def add_edges(self, edges):
        """Add edges between Labels."""
        self._graph.add_edges_from(edges)
        self.clear_topological_cache()

    def remove_node(self, node):
        """Remove node from the tree."""
        super().remove_node(node)
        self.clear_topological_cache()

    @property
    def num_labels(self):
        """Return the number of labels in the tree."""
        return self.num_nodes()

    def clear_topological_cache(self):
        """Clear the internal cache of the list of labels sorted in topological order.

        This function should be called if the topology of the graph has changed to
            prevent the cache from being stale.
        Note that it is automatically called when modifying the topology through the
            methods provided by this class.
        """
        self.__topological_order_cache = None

    def get_labels_in_topological_order(self) -> List[LabelEntity]:
        """Return a list of the labels in this graph sorted in topological order.

        To avoid performance issues, the output of this function is cached.

        Returns:
            List[LabelEntity]: sorted list of labels
        """
        if self.__topological_order_cache is None:
            # TODO: It seems that we are storing the edges the wrong way around.
            #       To work around this issue, we have to reverse the sorted list.
            self.__topological_order_cache = list(reversed(list(self.topological_sort())))

        return self.__topological_order_cache

    @property
    def type(self):
        """Returns the type of the LabelTree."""
        return "tree"

    def add_child(self, parent: LabelEntity, child: LabelEntity):
        """Add a `child` Label to `parent`."""
        self.add_edge(child, parent)
        self.clear_topological_cache()

    def get_parent(self, label: LabelEntity) -> Optional[LabelEntity]:
        """Returns the parent of `label`.

        Returns:
            the parent if it has one otherwise None
        """
        result = self.neighbors(label)
        return result[0] if len(result) > 0 else None

    def get_children(self, parent: LabelEntity) -> List[LabelEntity]:
        """Returns children of `parent`."""
        if parent not in self._graph.nodes:
            return []
        return list(self._graph.predecessors(parent))  # pylint: disable=no-member

    def get_descendants(self, parent: LabelEntity) -> List[LabelEntity]:
        """Returns descendants (children and children of children, etc.) of `parent`."""
        return self.descendants(parent)

    def get_siblings(self, label: LabelEntity) -> List[LabelEntity]:
        """Returns the siblings of a label."""
        parent = self.get_parent(label)
        if parent is None:
            siblings = []
        else:
            siblings = [u for u, v in self._graph.in_edges(parent) if u != label]  # pylint: disable=no-member
        return siblings

    def get_ancestors(self, label: LabelEntity) -> List[LabelEntity]:
        """Returns ancestors of `label`, including self."""
        result = []
        parent: Optional[LabelEntity] = label
        while parent is not None:
            result.append(parent)
            parent = self.get_parent(parent)
        return result

    def subgraph(self, labels: Sequence[LabelEntity]) -> "LabelTree":
        """Return the subgraph containing the given labels."""
        new_graph = LabelTree()
        new_graph.set_graph(self.get_graph().subgraph(labels).copy())
        return new_graph

    def __eq__(self, other) -> bool:
        """Check if two LabelTrees are equal."""
        if isinstance(other, LabelTree):
            return super().__eq__(other)
        return False


class LabelSchemaEntity:
    """This class represents the relationships of labels.

    This class currently keeps track of the following relationships:

    - parent/child label relationship
    - label group relationships

    Args:
        label_tree (LabelTree): a hierarchy of labels represented as a tree
        label_groups (List[LabelGroup]): list of groups of labels that form logical groups.
            E.g. a group of mutually exclusive labels.
    """

    # pylint: disable=too-many-public-methods, too-many-arguments
    def __init__(
        self,
        label_tree: LabelTree = None,
        label_groups: List[LabelGroup] = None,
    ):
        if label_tree is None:
            label_tree = LabelTree()
        self.label_tree = label_tree

        if label_groups is None:
            label_groups = []
        self._groups = label_groups

    def get_labels(self, include_empty: bool) -> List[LabelEntity]:
        """Get the labels in the label schema.

        Args:
            include_empty (bool): flag determining whether to include empty
                labels

        Returns:
            List[LabelEntity]: list of all labels in the label schema
        """
        labels = {label for group in self._groups for label in group.labels if include_empty or not label.is_empty}
        return sorted(list(labels), key=natural_sort_label_id)

    def get_groups(self, include_empty: bool = False) -> List[LabelGroup]:
        """Get the label groups in the label schema.

        Args:
            include_empty (bool): flag determining whether to include empty
                label groups

        Returns:
            List[LabelGroup]: list of all label groups in the label schema
        """
        if include_empty:
            return self._groups

        return [group for group in self._groups if group.group_type != LabelGroupType.EMPTY_LABEL]

    def add_group(self, label_group: LabelGroup):
        """Adding a group to label schema.

        Args:
            label_group (LabelGroup): label group to add

        Returns:
            None
        """
        if label_group.name in [group.name for group in self._groups]:
            raise LabelGroupExistsException(
                f"group with '{label_group.name}' exists, " f"use add_labels_to_group_by_group_name instead"
            )
        self.__append_group(label_group)

    def add_child(self, parent: LabelEntity, child: LabelEntity):
        """Add a `child` Label to `parent`."""
        parent = self.__get_label(parent)
        child = self.__get_label(child)
        self.label_tree.add_child(parent, child)

    def get_parent(self, label: LabelEntity) -> Optional[LabelEntity]:
        """Returns the parent of `label`.

        Returns:
            Optional[LabelEntity]: the parent if it has one otherwise None
        """
        label = self.__get_label(label)
        return self.label_tree.get_parent(label)

    def get_label_ids(self, include_empty: bool) -> List[ID]:
        """Returns a list of label ids that are in the LabelSchema.

        Args:
            include_empty (bool): Include empty label id or not.

        Returns:
            List[ID]: sorted list of label ids
        """
        label_ids = {
            label.id_ for group in self._groups for label in group.labels if include_empty or not label.is_empty
        }
        return sorted(list(label_ids), key=natural_sort_label_id)

    def get_label_group_by_name(self, group_name: str) -> Optional[LabelGroup]:
        """Get the label group by the passed group_name.

        Args:
            group_name (str): name of the group to get

        Returns:
            Optional[LabelGroup]
        """
        for label_group in self._groups:
            if group_name == label_group.name:
                return label_group
        return None

    def get_exclusive_groups(self) -> List[LabelGroup]:
        """Returns exclusive groups in the LabelSchema."""

        return [group for group in self._groups if group.group_type == LabelGroupType.EXCLUSIVE]

    def add_labels_to_group_by_group_name(self, group_name: str, labels: Sequence[LabelEntity]):
        """Adds `labels` to group named `group_name`.

        Args:
            labels (str): list of Label
            group_name (Sequence[LabelEntity]): group name

        Raises:
            LabelGroupDoesNotExistException: This is raised if the group
                does not exist
        """
        group = self.get_label_group_by_name(group_name)
        if group is not None:
            group.labels.extend(labels)
        else:
            raise LabelGroupDoesNotExistException(f"group with name '{group_name}' does not exist, cannot add")

    def __append_group(self, label_group: LabelGroup):
        """Convenience function for appending `label_group` to the necessary internal data structures.

        Args:
            label_group (LabelGroup): label group to append
        """
        if label_group not in self._groups:
            self._groups.append(label_group)

    def are_exclusive(self, label1: LabelEntity, label2: LabelEntity) -> bool:
        """Returns whether `label` and `label2` are mutually exclusive."""
        return label2 in self.get_labels_exclusive_to(label1)

    def get_children(self, parent: LabelEntity) -> List[LabelEntity]:
        """Return a list of the children of the passed parent Label."""
        parent = self.__get_label(parent)
        return self.label_tree.get_children(parent)

    def get_siblings_in_group(self, label: LabelEntity) -> List[LabelEntity]:
        """Return a list of the 'siblings', which are all labels within the same group as a label."""
        containing_group = self.get_group_containing_label(label)
        if containing_group is None:
            return []
        return [label_iter for label_iter in containing_group.labels if not label_iter == label]

    def get_descendants(self, parent: LabelEntity) -> List[LabelEntity]:
        """Returns descendants (children and children of children, etc.) of `parent`."""
        parent = self.__get_label(parent)
        return self.label_tree.get_descendants(parent)

    def get_ancestors(self, label: LabelEntity) -> List[LabelEntity]:
        """Returns ancestors of `label`, including self."""
        label = self.__get_label(label)
        return self.label_tree.get_ancestors(label)

    def get_group_containing_label(self, label: LabelEntity) -> Optional[LabelGroup]:
        """Returns the label group which contains the label.

        Args:
            label (LabelEntity): the query label

        Returns:
            Optional[LabelGroup]: the group containing the label
        """
        label = self.__get_label(label)
        for group in self._groups:
            if label in group.labels:
                return group
        return None

    def get_labels_exclusive_to(self, label: LabelEntity) -> List[LabelEntity]:
        """Returns a list of labels that are exclusive to the passed label."""
        if label.is_empty:
            exclusive_labels = self.__get_exclusivity_for_empty_label(label=label)
        else:
            exclusive_labels = self.__get_exclusivity_recursion(label=label)
        return exclusive_labels

    def __get_exclusivity_recursion(self, label: LabelEntity, add_empty: bool = True) -> List[LabelEntity]:
        """Recursively computes all labels exclusive to a non-empty label.

        A label is exclusive with:
            - All labels in the same group
            - All children of labels in the same group
            - All labels in the same group as any of the label's ancestors
            - All children of labels in the same group as any of the label's ancestors
            - All empty labels that are not descendants of the label

        Args:
            label (LabelEntity): The label to get exclusive labels for
            add_empty (bool): If set to True, adds all empty labels that are
                not descendants of the label. This is only
        needed for the first recursion iteration.

        Returns:
            List[LabelEntity]: List of labels exclusive to the label.
        """
        output = []

        # Add all labels in the same group
        siblings = self.get_siblings_in_group(label)
        output += siblings

        # Add all children of labels in the same group
        for sibling in siblings:
            output += self.get_children(sibling)

        # Do the same for the parent of the label
        parent = self.get_parent(label)
        if parent is not None:
            output += self.__get_exclusivity_recursion(parent, add_empty=False)

        # Add all empty labels that are not descendants of the label. We don't need to check the ancestors for being
        # empty, because empty label's don't have descendants.
        if add_empty:
            descendants = self.get_descendants(label)
            exclusive_empty_labels = [
                label_iter
                for label_iter in self.get_labels(include_empty=True)
                if label_iter.is_empty and label_iter not in descendants
            ]
            output = list(set(output + exclusive_empty_labels))
        return output

    def __get_exclusivity_for_empty_label(self, label: LabelEntity) -> List[LabelEntity]:
        """Get the labels exclusive to an empty label.

        For an empty label, all labels are exclusive to it except it's ancestors.

        Args:
            label (LabelEntity): empty Label to get exclusive labels for

        Returns:
            List[LabelEntity]: List of Labels exclusive to the Label
        """
        ancestors = self.get_ancestors(label)
        return [label for label in self.get_labels(include_empty=True) if label not in ancestors]

    @staticmethod
    def __get_label(label: Union[ScoredLabel, LabelEntity]) -> LabelEntity:
        """Returns Label object from possibly non-label object.

        Args:
            label (Union[ScoredLabel, LabelEntity]): label to get Label object for

        Returns:
            LabelEntity: Label object
        """
        if isinstance(label, ScoredLabel):
            returned_label = label.get_label()
        elif isinstance(label, LabelEntity):
            returned_label = label
        else:
            raise ValueError("Input of __get_label is not of type Label or ScoredLabel")
        return returned_label  # type: ignore

    def __repr__(self) -> str:
        """String representation of LabelSchemaEntity."""
        return f"LabelSchemaEntity(label_groups={self._groups})"

    def __eq__(self, other) -> bool:
        """Returns whether two LabelSchemaEntities are equal."""
        if isinstance(other, LabelSchemaEntity):
            return self.label_tree == other.label_tree and self.get_groups(include_empty=True) == other.get_groups(
                include_empty=True
            )
        return False

    @classmethod
    def from_labels(cls, labels: Sequence[LabelEntity]) -> "LabelSchemaEntity":
        """Create LabelSchemaEntity from a list of exclusive labels.

        Args:
            labels (Sequence[LabelEntity]): list of labels

        Returns:
            LabelSchemaEntity from the given labels
        """
        label_group = LabelGroup(name="from_label_list", labels=labels)
        return LabelSchemaEntity(label_groups=[label_group])

    def resolve_labels_probabilistic(
        self,
        scored_labels: List[ScoredLabel],
        selected_labels: List[LabelEntity] = None,
    ) -> List[ScoredLabel]:
        """Resolves hierarchical labels and exclusivity based on a list of ScoredLabels (labels with probability).

        The following two steps are taken:

        - selects the most likely label from an exclusive (multiclass) group
        - removes children of "not-most-likely" (non-max) parents in an exclusive group (top-down approach)

        The method is intended to post-process the output of probabilistic systems such as predictions coming from
        machine learning methods to resolve ambiguities and logical impossibilities. When processing (non-probabilistic)
        user input please use `complete_labels` instead.

        Args:
            scored_labels (List[ScoredLabel]): a list of ScoredLabels (labels with
                probability)
            selected_labels (List[LabelEntity]): if not None, will only consider labels
                within `selected_labels` for resolving. Any other labels
                which have relations with selected_labels (e.g. parent),
                but are outside `selected_labels` are set to a default
                probability of 1.0
        """
        input_domains = set(lbl.domain for lbl in scored_labels)
        label_to_probability = {scored_label.get_label(): scored_label.probability for scored_label in scored_labels}
        resolved_labels = self.__resolve_labels_probabilistic(label_to_probability, selected_labels)
        output_domains = set(lbl.domain for lbl in resolved_labels)
        if input_domains != output_domains:
            logger.error(
                "Something went wrong in 'resolve_labels_probabilistic', "
                "some tasks (domains) lost all their labels; "
                "label_schema: %s  input_labels: %s  output_labels: %s",
                self,
                scored_labels,
                resolved_labels,
            )
        return resolved_labels

    def __resolve_labels_probabilistic(
        self,
        label_to_probability: Dict[LabelEntity, float],
        selected_labels: Optional[Sequence[LabelEntity]],
    ) -> List[ScoredLabel]:
        """Resolves hierarchical labels and exclusivity based on a probabilistic label output.

        - selects the most likely (max) label from an exclusive group
        - removes children of non-max parents in an exclusive group

        See `resolve_labels_probabilistic` for parameter descriptions

        Args:
            label_to_probability (Dict[LabelEntity, float]): map from `Label` to float.
            selected_labels (Optional[Sequence[LabelEntity]]): Subset of labels.

        Returns:
            List[ScoredLabel]: List of ScoredLabels (labels with probability)
        """
        # add (potentially) missing ancestors labels for children with probability 0
        # this is needed so that suppression of children of non-max exclusive labels works when the exclusive
        # group has only one member
        label_to_probability = self.__add_missing_ancestors(label_to_probability, selected_labels)

        hard_classification = self.__resolve_exclusive_labels(label_to_probability)

        # suppress the output of children of parent nodes that are not the most likely label within their group
        resolved = self.__suppress_descendant_output(hard_classification)

        result = []
        for label, probability in resolved.items():
            if probability > 0:  # only return labels with non-zero probability
                result.append(
                    ScoredLabel(
                        label,
                        probability=(
                            probability
                            * label_to_probability.get(label, 1.0)
                            # retain the original probability in the output
                        ),
                    )
                )
        return result

    def __suppress_descendant_output(self, hard_classification: Dict[LabelEntity, float]) -> Dict[LabelEntity, float]:
        """Suppresses outputs in `label_to_probability`.

        Sets probability to 0.0 for descendants of parents that have 0 probability in `hard_classification`.
        """

        # Input: Conditional probability of each label given its parent label
        # Output: Marginal probability of each label

        # We recursively compute the marginal probability of each node by multiplying the conditional probability
        # with the marginal probability of its parent. That is:
        # P(B) = P(B|A) * P(A)
        # The recursion is done a topologically sorted list of labels to ensure that the marginal probability
        # of the parent label has been computed before trying to compute the child's probability.

        label_tree = self.label_tree
        all_labels = label_tree.get_labels_in_topological_order()

        for child in all_labels:
            if child in hard_classification:
                # Get the immediate parents (should be at most one element; zero for root labels)
                parents = label_tree.neighbors(child)

                if len(parents) > 0:
                    parent = parents[0]
                    if parent in hard_classification:
                        hard_classification[child] *= hard_classification[parent]

        return hard_classification

    def __resolve_exclusive_labels(self, label_to_probability: Dict[LabelEntity, float]) -> Dict[LabelEntity, float]:
        """Resolve exclusive labels.

        For labels in `label_to_probability` sets labels that are most likely (maximum probability) in their exclusive
        group to 1.0 and other (non-max) labels to probability 0.
        """
        hard_classification: Dict[LabelEntity, float] = {}
        top_level_labels_in_label_schema = [
            label_ for label_ in self.label_tree.get_labels_in_topological_order() if self.get_parent(label_) is None
        ]

        for label, probability in label_to_probability.items():
            if label not in hard_classification:
                label_parent = self.get_parent(label)
                if label_parent is None:
                    # The label itself is a top-level label
                    exclusive_neighbours = [label_ for label_ in top_level_labels_in_label_schema if label_ != label]
                else:
                    exclusive_neighbours = [label_ for label_ in self.get_children(label_parent) if label_ != label]

                probabilities = [probability]
                neighbours_ = [label]
                for neighbor in exclusive_neighbours:
                    neighbours_.append(neighbor)
                    probabilities.append(label_to_probability.get(neighbor, 0))
                if len(probabilities) > 1:
                    max_index = np.argmax(probabilities)
                    for idx, neighbor in enumerate(neighbours_):
                        hard_classification[neighbor] = float(max_index == idx)
                else:
                    # single node group, interpret as multilabel node
                    hard_classification[label] = float(label_to_probability[label] > 0.0)
        return hard_classification

    def __add_missing_ancestors(
        self,
        label_to_probability: Dict[LabelEntity, float],
        selected_labels: Optional[Sequence[LabelEntity]],
    ) -> Dict[LabelEntity, float]:
        """Adds missing ancestors (of the same task) to the `label_to_probability` map.

        Missing ancestors get probability `probability`
        """
        updated_label_to_probability = copy.deepcopy(label_to_probability)
        for label in label_to_probability:
            for ancestor in self.get_ancestors(label):
                if ancestor not in updated_label_to_probability:
                    updated_label_to_probability[ancestor] = (
                        0.0  # by default missing ancestors get probability 0.0
                        if selected_labels is None
                        else (ancestor not in selected_labels) * 1.0
                        # ... unless label selection is used, in that case
                        # the ancestor will get probability 1.0 if it is missing
                    )
        return updated_label_to_probability



"""Train Task of OTX Classification."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=invalid-name

import io
from typing import List, Optional

import torch
from mmcv.utils import ConfigDict

from otx.algorithms.classification.configs import ClassificationConfig
from otx.algorithms.common.utils.callback import TrainingProgressCallback
from otx.algorithms.common.utils.data import get_dataset
from otx.api.configuration import cfg_helper
from otx.api.configuration.helper.utils import ids_to_strings
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.metrics import (
    CurveMetric,
    LineChartInfo,
    LineMetricsGroup,
    MetricsGroup,
    Performance,
    ScoreMetric,
)
from otx.api.entities.model import ModelEntity  # ModelStatus
from otx.api.entities.subset import Subset
from otx.api.entities.train_parameters import TrainParameters
from otx.api.entities.train_parameters import (
    default_progress_callback as train_default_progress_callback,
)
from otx.api.serialization.label_mapper import label_schema_to_bytes
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    check_input_parameters_type,
)
from otx.mpa.utils.logger import get_logger

from .inference import ClassificationInferenceTask

logger = get_logger()

TASK_CONFIG = ClassificationConfig


# pylint: disable= too-many-ancestors
class ClassificationTrainTask(ClassificationInferenceTask):
    """Train Task Implementation of OTX Classification."""

    @check_input_parameters_type()
    def save_model(self, output_model: ModelEntity):
        """Save best model weights in ClassificationTrainTask."""
        logger.info("called save_model")
        buffer = io.BytesIO()
        hyperparams_str = ids_to_strings(cfg_helper.convert(self._hyperparams, dict, enum_to_str=True))
        labels = {label.name: label.color.rgb_tuple for label in self._labels}
        model_ckpt = torch.load(self._model_ckpt)
        modelinfo = {
            "model": model_ckpt,
            "config": hyperparams_str,
            "labels": labels,
            "VERSION": 1,
        }

        torch.save(modelinfo, buffer)
        output_model.set_data("weights.pth", buffer.getvalue())
        output_model.set_data(
            "label_schema.json",
            label_schema_to_bytes(self._task_environment.label_schema),
        )
        output_model.precision = self._precision

    def cancel_training(self):
        """Cancel training function in ClassificationTrainTask.

        Sends a cancel training signal to gracefully stop the optimizer.
        The signal consists of creating a '.stop_training' file in the current work_dir.
        The runner checks for this file periodically.
        The stopping mechanism allows stopping after each iteration, but validation will still be carried out.
        Stopping will therefore take some time.
        """
        self._should_stop = True
        logger.info("Cancel training requested.")
        if self.cancel_interface is not None:
            self.cancel_interface.cancel()
        else:
            logger.info("but training was not started yet. reserved it to cancel")
            self.reserved_cancel = True

    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def train(
        self,
        dataset: DatasetEntity,
        output_model: ModelEntity,
        train_parameters: Optional[TrainParameters] = None,
    ):
        """Train function in ClassificationTrainTask."""
        logger.info("train()")
        # Check for stop signal between pre-eval and training.
        # If training is cancelled at this point,
        if self._should_stop:
            logger.info("Training cancelled.")
            self._should_stop = False
            self._is_training = False
            return

        # Set OTX LoggerHook & Time Monitor
        update_progress_callback = train_default_progress_callback
        if train_parameters is not None:
            update_progress_callback = train_parameters.update_progress  # type: ignore
        self._time_monitor = TrainingProgressCallback(update_progress_callback)

        stage_module = "ClsTrainer"
        self._data_cfg = self._init_train_data_cfg(dataset)
        self._is_training = True
        results = self._run_task(stage_module, mode="train", dataset=dataset, parameters=train_parameters)

        # Check for stop signal between pre-eval and training.
        # If training is cancelled at this point,
        if self._should_stop:
            logger.info("Training cancelled.")
            self._should_stop = False
            self._is_training = False
            return

        # get output model
        model_ckpt = results.get("final_ckpt")
        if model_ckpt is None:
            logger.error("cannot find final checkpoint from the results.")
            return
        # update checkpoint to the newly trained model
        self._model_ckpt = model_ckpt

        # compose performance statistics
        training_metrics, final_acc = self._generate_training_metrics_group(self._learning_curves)
        # save resulting model
        self.save_model(output_model)
        performance = Performance(
            score=ScoreMetric(value=final_acc, name="accuracy"),
            dashboard_metrics=training_metrics,
        )
        logger.info(f"Final model performance: {str(performance)}")
        output_model.performance = performance
        self._is_training = False
        logger.info("train done.")

    def _init_train_data_cfg(self, dataset: DatasetEntity):
        logger.info("init data cfg.")
        data_cfg = ConfigDict(data=ConfigDict())

        for cfg_key, subset in zip(
            ["train", "val", "unlabeled"],
            [Subset.TRAINING, Subset.VALIDATION, Subset.UNLABELED],
        ):
            subset = get_dataset(dataset, subset)
            if subset:
                data_cfg.data[cfg_key] = ConfigDict(
                    otx_dataset=subset,
                    labels=self._labels,
                )

        for label in self._labels:
            label.hotkey = "a"
        return data_cfg

    def _generate_training_metrics_group(self, learning_curves):
        """Parses the classification logs to get metrics from the latest training run.

        :return output List[MetricsGroup]
        """
        output: List[MetricsGroup] = []

        if self._multilabel:
            metric_key = "val/accuracy-mlc"
        elif self._hierarchical:
            metric_key = "val/MHAcc"
        else:
            metric_key = "val/accuracy_top-1"

        # Learning curves
        best_acc = -1
        if learning_curves is None:
            return output

        for key, curve in learning_curves.items():
            metric_curve = CurveMetric(xs=curve.x, ys=curve.y, name=key)
            if key == metric_key:
                best_acc = max(curve.y)
            visualization_info = LineChartInfo(name=key, x_axis_label="Timestamp", y_axis_label=key)
            output.append(LineMetricsGroup(metrics=[metric_curve], visualization_info=visualization_info))

        return output, best_acc



_base_ = "./encoder_decoder.ote.py"

__norm_cfg = dict(type="BN", requires_grad=True)
model = dict(
    type="ClassIncrEncoderDecoder",
    decode_head=dict(
        type="FCNHead",
        in_channels=40,
        in_index=0,
        channels=40,
        input_transform=None,
        kernel_size=1,
        num_convs=0,
        concat_input=False,
        dropout_ratio=-1,
        num_classes=19,
        norm_cfg=__norm_cfg,
        align_corners=False,
        enable_out_norm=False,
        loss_decode=[
            dict(
                type="CrossEntropyLoss",
                use_sigmoid=False,
                loss_weight=1.0,
            )
        ],
    ),
    test_cfg=dict(
        mode="whole",
        output_scale=3.0,
    ),
)



import os

import mmcv
import pytest

from otx.mpa.stage import Stage, get_available_types
from tests.test_suite.e2e_test_system import e2e_pytest_unit


@e2e_pytest_unit
def test_get_available_types():
    return_value = get_available_types()
    assert isinstance(return_value, list)


class TestStage:
    @e2e_pytest_unit
    def test_init(self, mocker):
        fake_cfg = {
            "work_dir": "test_init",
            "total_epochs": 5,
            "runner": {"max_epochs": 10},
            "checkpoint_config": {"interval": 20},
            "seed": 0,
        }
        fake_common_cfg = {"output_path": "/path/output"}
        mocker.patch.object(mmcv, "mkdir_or_exist")
        stage = Stage("mpa_test", "", fake_cfg, fake_common_cfg, 0, fake_kwargs=None)

        assert stage.cfg.get("runner", False)
        assert stage.cfg.runner.get("max_epochs", 5)
        assert stage.cfg.get("checkpoint_config", False)
        assert stage.cfg.checkpoint_config.get("interval", 5)

        work_dir = str(stage.cfg.get("work_dir", ""))

        assert work_dir.find("test_init")
        assert work_dir.find("mpa_test")

    @e2e_pytest_unit
    def test_init_with_distributed_enabled(self, mocker):
        fake_cfg = {"work_dir": "test_init_distributed"}
        fake_common_cfg = {"output_path": "/path/output"}
        mocker.patch.object(mmcv, "mkdir_or_exist")
        mocker.patch("torch.distributed.is_initialized", returned_value=True)
        os.environ["LOCAL_RANK"] = "2"
        stage = Stage("mpa_test", "", fake_cfg, fake_common_cfg, 0)

        assert stage.distributed is True

    @e2e_pytest_unit
    def test_configure_data(self):
        _base_pipeline = [dict(type="LoadImageFromFile"), dict(type="MultiScaleFlipAug")]
        _transform = [
            dict(type="Resize", keep_ratio=False),
            dict(type="ImageToTensor", keys=["img"]),
            dict(type="Collect", keys=["img"]),
        ]
        data_cfg = mmcv.ConfigDict(
            data=dict(
                train=dict(dataset=dict(pipeline=_base_pipeline)),
                val=dict(pipeline=_base_pipeline),
                pipeline_options=dict(
                    MultiScaleFlipAug=dict(img_scale=(224, 224), transforms=_transform),
                ),
            )
        )
        Stage.configure_data(data_cfg, True)

        assert data_cfg.data.train.dataset.pipeline[1].img_scale == (224, 224)
        assert data_cfg.data.train.dataset.pipeline[1].transforms == _transform
        assert data_cfg.data.val.pipeline[1].img_scale == (224, 224)
        assert data_cfg.data.val.pipeline[1].transforms == _transform

    @e2e_pytest_unit
    def test_configure_ckpt(self, mocker):
        cfg = mmcv.ConfigDict(load_from=None)
        mocker.patch.object(mmcv, "mkdir_or_exist")
        stage = Stage("mpa_test", "", {}, {}, 0)

        mocker.patch.object(stage, "get_model_ckpt", return_value="/path/to/load/ckpt")
        stage.configure_ckpt(cfg, "/foo/bar")

        assert cfg.load_from == "/path/to/load/ckpt"

    @e2e_pytest_unit
    def test_configure_hook(self):
        cfg = mmcv.ConfigDict(
            custom_hook_options=dict(MockHook=dict(opt1="MockOpt1", opt2="MockOpt2")),
            custom_hooks=[dict(type="MockHook")],
        )
        Stage.configure_hook(cfg)

        assert cfg.custom_hooks[0].opt1 == "MockOpt1"
        assert cfg.custom_hooks[0].opt2 == "MockOpt2"

    @e2e_pytest_unit
    def test_configure_samples_per_gpu(self, mocker):
        cfg = mmcv.ConfigDict(data=dict(train_dataloader=dict(samples_per_gpu=2)))
        mock_otx_dataset = mocker.MagicMock()
        mock_otx_dataset.__len__.return_value = 1

        mocker.patch("otx.mpa.stage.get_data_cfg", return_value=mmcv.ConfigDict(otx_dataset=mock_otx_dataset))
        Stage.configure_samples_per_gpu(cfg, "train", False)

        assert "train_dataloader" in cfg.data
        assert cfg.data["train_dataloader"]["samples_per_gpu"] == 1

    @e2e_pytest_unit
    def test_configure_compat_cfg(self):
        cfg = mmcv.ConfigDict(
            data=dict(
                train=None,
                val=None,
                test=None,
                unlabeled=None,
            )
        )
        with pytest.raises(AttributeError):
            Stage.configure_compat_cfg(cfg)

    @e2e_pytest_unit
    def test_configure_fp16_optimizer(self):
        cfg = mmcv.ConfigDict(fp16=dict(loss_scale=512.0), optimizer_config=dict(type="OptimizerHook"))

        Stage.configure_fp16_optimizer(cfg)
        assert cfg.optimizer_config.type == "Fp16OptimizerHook"

    @e2e_pytest_unit
    def test_configure_unlabeled_dataloader(self, mocker):
        cfg = mmcv.ConfigDict(
            data=dict(
                unlabeled=dict(),
            ),
            model_task="classification",
        )

        mocker.patch("importlib.import_module")
        mock_build_ul_dataset = mocker.patch("otx.mpa.stage.build_dataset")
        mock_build_ul_dataloader = mocker.patch("otx.mpa.stage.build_dataloader")

        Stage.configure_unlabeled_dataloader(cfg)

        mock_build_ul_dataset.assert_called_once()
        mock_build_ul_dataloader.assert_called_once()
        assert cfg.custom_hooks[0].type == "ComposedDataLoadersHook"

    @e2e_pytest_unit
    def test_get_model_meta(self, mocker):
        cfg = dict(load_from="/foo/bar")
        from mmcv.runner import CheckpointLoader

        mock_load_ckpt = mocker.patch.object(
            CheckpointLoader, "load_checkpoint", return_value={"meta": {"model_meta": None}}
        )

        returned_value = Stage.get_model_meta(cfg)

        mock_load_ckpt.assert_called_once()
        assert returned_value == {"model_meta": None}

    @e2e_pytest_unit
    def test_get_data_cfg(self, mocker):
        cfg = mmcv.ConfigDict(data=dict(train=dict(dataset=dict(dataset="config"))))
        returned_value = Stage.get_data_cfg(cfg, "train")

        assert returned_value == "config"

    @e2e_pytest_unit
    def test_get_data_classes(self, mocker):
        cfg = mmcv.ConfigDict()
        mocker.patch.object(Stage, "get_data_cfg", return_value={"data_classes": ["foo"]})

        returned_value = Stage.get_data_classes(cfg)

        assert returned_value == ["foo"]

    @e2e_pytest_unit
    def test_get_model_classes(self, mocker):
        cfg = mmcv.ConfigDict(model=dict())
        mocker.patch.object(Stage, "get_model_meta", return_value={"CLASSES": ["foo", "bar"]})
        mocker.patch.object(Stage, "read_label_schema")
        returned_value = Stage.get_model_classes(cfg)

        assert returned_value == ["foo", "bar"]

    @e2e_pytest_unit
    def test_get_model_ckpt(self, mocker):
        from mmcv.runner import CheckpointLoader

        mock_load_ckpt = mocker.patch.object(
            CheckpointLoader, "load_checkpoint", return_value=dict(model=mocker.MagicMock())
        )
        mock_save = mocker.patch("torch.save")

        Stage.get_model_ckpt("/ckpt/path/weights.pth")

        mock_load_ckpt.assert_called_once()
        mock_save.assert_called_once()

    @e2e_pytest_unit
    def test_read_label_schema(self, mocker):
        mocker.patch("os.path.exists", return_value=True)
        mocker.patch("builtins.open")
        mocker.patch("json.load", return_value=dict(all_labels=dict(label=dict(name="foo"))))
        returned_value = Stage.read_label_schema("/ckpt/path/weights.pth")

        assert returned_value == ["foo"]

    @e2e_pytest_unit
    def test_set_inference_progress_callback(self, mocker):
        mock_model = mocker.MagicMock()
        mock_model.register_forward_pre_hook()
        mock_model.register_forward_hook()
        cfg = dict(custom_hooks=[dict(name="OTXProgressHook", time_monitor=mocker.MagicMock())])

        Stage.set_inference_progress_callback(cfg, mock_model)

        mock_model.register_forward_pre_hook.assert_called_once()
        mock_model.register_forward_hook.assert_called_once()

    @e2e_pytest_unit
    def test_build_model(self, mocker):
        mock_model_builder = mocker.patch.object(Stage, "MODEL_BUILDER")
        Stage.build_model(cfg=mmcv.Config())

        mock_model_builder.assert_called_once()



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch.nn as nn
from mmseg.core import add_prefix
from mmseg.models.builder import build_loss
from mmseg.ops import resize

from ..losses.utils import LossEqualizer


class PixelWeightsMixin(object):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._init_train_components(self.train_cfg)
        self.feature_maps = None

    def _init_train_components(self, train_cfg):
        if train_cfg is None:
            self.mutual_losses = None
            self.loss_equalizer = None
            return

        mutual_loss_configs = train_cfg.get("mutual_loss")
        if mutual_loss_configs:
            if isinstance(mutual_loss_configs, dict):
                mutual_loss_configs = [mutual_loss_configs]

            self.mutual_losses = nn.ModuleList()
            for mutual_loss_config in mutual_loss_configs:
                self.mutual_losses.append(build_loss(mutual_loss_config))
        else:
            self.mutual_losses = None

        loss_reweighting_config = train_cfg.get("loss_reweighting")
        if loss_reweighting_config:
            self.loss_equalizer = LossEqualizer(**loss_reweighting_config)
        else:
            self.loss_equalizer = None

    @staticmethod
    def _get_argument_by_name(trg_name, arguments):
        assert trg_name in arguments.keys()
        return arguments[trg_name]

    def set_step_params(self, init_iter, epoch_size):
        self.decode_head.set_step_params(init_iter, epoch_size)

        if getattr(self, "auxiliary_head", None) is not None:
            if isinstance(self.auxiliary_head, nn.ModuleList):
                for aux_head in self.auxiliary_head:
                    aux_head.set_step_params(init_iter, epoch_size)
            else:
                self.auxiliary_head.set_step_params(init_iter, epoch_size)

    def _decode_head_forward_train(self, x, img_metas, pixel_weights=None, **kwargs):

        trg_map = self._get_argument_by_name(self.decode_head.loss_target_name, kwargs)
        loss_decode, logits_decode = self.decode_head.forward_train(
            x,
            img_metas,
            trg_map,
            train_cfg=self.train_cfg,
            pixel_weights=pixel_weights,
            return_logits=True,
        )

        scale = self.decode_head.last_scale
        scaled_logits_decode = scale * logits_decode

        name_prefix = "decode"

        losses, meta = dict(), dict()
        losses.update(add_prefix(loss_decode, name_prefix))
        meta[f"{name_prefix}_scaled_logits"] = scaled_logits_decode

        return losses, meta

    def _auxiliary_head_forward_train(self, x, img_metas, **kwargs):

        losses, meta = dict(), dict()
        if isinstance(self.auxiliary_head, nn.ModuleList):
            for idx, aux_head in enumerate(self.auxiliary_head):
                trg_map = self._get_argument_by_name(aux_head.loss_target_name, kwargs)
                loss_aux, logits_aux = aux_head.forward_train(
                    x,
                    img_metas,
                    trg_map,
                    train_cfg=self.train_cfg,
                    return_logits=True,
                )

                scale = aux_head.last_scale
                scaled_logits_aux = scale * logits_aux

                name_prefix = f"aux_{idx}"
                losses.update(add_prefix(loss_aux, name_prefix))
                meta[f"{name_prefix}_scaled_logits"] = scaled_logits_aux
        else:
            trg_map = self._get_argument_by_name(self.auxiliary_head.loss_target_name, kwargs)
            loss_aux, logits_aux = self.auxiliary_head.forward_train(
                x,
                img_metas,
                trg_map,
                train_cfg=self.train_cfg,
                return_logits=True,
            )

            scale = self.auxiliary_head.last_scale
            scaled_logits_aux = scale * logits_aux

            name_prefix = "aux"
            losses.update(add_prefix(loss_aux, name_prefix))
            meta[f"{name_prefix}_scaled_logits"] = scaled_logits_aux

        return losses, meta

    def forward_train(self, img, img_metas, gt_semantic_seg, pixel_weights=None, **kwargs):
        """Forward function for training.

        Args:
            img (Tensor): Input images.
            img_metas (list[dict]): List of image info dict where each dict
                has: 'img_shape', 'scale_factor', 'flip', and may also contain
                'filename', 'ori_shape', 'pad_shape', and 'img_norm_cfg'.
                For details on the values of these keys see
                `mmseg/datasets/pipelines/formatting.py:Collect`.
            gt_semantic_seg (Tensor): Semantic segmentation masks
                used if the architecture supports semantic segmentation task.
            pixel_weights (Tensor): Pixels weights.

        Returns:
            dict[str, Tensor]: a dictionary of loss components
        """

        losses = dict()

        features = self.extract_feat(img)

        loss_decode, meta_decode = self._decode_head_forward_train(
            features, img_metas, pixel_weights, gt_semantic_seg=gt_semantic_seg, **kwargs
        )
        losses.update(loss_decode)

        if self.with_auxiliary_head:
            loss_aux, meta_aux = self._auxiliary_head_forward_train(
                features, img_metas, gt_semantic_seg=gt_semantic_seg, **kwargs
            )
            losses.update(loss_aux)

        if self.mutual_losses is not None and self.with_auxiliary_head:
            meta = dict()
            meta.update(meta_decode)
            meta.update(meta_aux)

            out_mutual_losses = dict()
            for mutual_loss_idx, mutual_loss in enumerate(self.mutual_losses):
                logits_a = self._get_argument_by_name(mutual_loss.trg_a_name, meta)
                logits_b = self._get_argument_by_name(mutual_loss.trg_b_name, meta)

                logits_a = resize(
                    input=logits_a, size=gt_semantic_seg.shape[2:], mode="bilinear", align_corners=self.align_corners
                )
                logits_b = resize(
                    input=logits_b, size=gt_semantic_seg.shape[2:], mode="bilinear", align_corners=self.align_corners
                )

                mutual_labels = gt_semantic_seg.squeeze(1)
                mutual_loss_value, mutual_loss_meta = mutual_loss(logits_a, logits_b, mutual_labels)

                mutual_loss_name = mutual_loss.name + f"-{mutual_loss_idx}"
                out_mutual_losses[mutual_loss_name] = mutual_loss_value
                losses[mutual_loss_name] = mutual_loss_value
                losses.update(add_prefix(mutual_loss_meta, mutual_loss_name))

            losses["loss_mutual"] = sum(out_mutual_losses.values())

        if self.loss_equalizer is not None:
            unweighted_losses = {loss_name: loss for loss_name, loss in losses.items() if "loss" in loss_name}
            weighted_losses = self.loss_equalizer.reweight(unweighted_losses)

            for loss_name, loss_value in weighted_losses.items():
                losses[loss_name] = loss_value

        return losses



"""Adapters for mmcv support."""

# Copyright (C) 2021-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from .hooks import (
    CancelTrainingHook,
    EarlyStoppingHook,
    EMAMomentumUpdateHook,
    EnsureCorrectBestCheckpointHook,
    OTXLoggerHook,
    OTXProgressHook,
    ReduceLROnPlateauLrUpdaterHook,
    StopLossNanTrainingHook,
    TwoCropTransformHook,
)
from .nncf.hooks import CompressionHook
from .nncf.runners import AccuracyAwareRunner
from .runner import EpochRunnerWithCancel, IterBasedRunnerWithCancel

__all__ = [
    "EpochRunnerWithCancel",
    "IterBasedRunnerWithCancel",
    "CancelTrainingHook",
    "OTXLoggerHook",
    "OTXProgressHook",
    "EarlyStoppingHook",
    "ReduceLROnPlateauLrUpdaterHook",
    "EnsureCorrectBestCheckpointHook",
    "StopLossNanTrainingHook",
    "EMAMomentumUpdateHook",
    "CompressionHook",
    "AccuracyAwareRunner",
    "TwoCropTransformHook",
]



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Dict, List, Optional, Union

import openvino.runtime as ov
import torch.nn.functional as F
from mmcls.models.builder import HEADS
from mmcls.models.heads import ClsHead

from ....graph.parsers.cls import cls_base_parser
from ...mmov_model import MMOVModel


@HEADS.register_module()
class MMOVClsHead(ClsHead):
    def __init__(
        self,
        model_path_or_model: Union[str, ov.Model],
        weight_path: Optional[str] = None,
        inputs: Optional[Union[Dict[str, Union[str, List[str]]], List[str], str]] = None,
        outputs: Optional[Union[Dict[str, Union[str, List[str]]], List[str], str]] = None,
        init_weight: bool = False,
        verify_shape: bool = True,
        softmax_at_test: bool = True,
        **kwargs,
    ):
        kwargs.pop("in_channels", None)
        kwargs.pop("num_classes", None)
        super().__init__(**kwargs)

        self._model_path_or_model = model_path_or_model
        self._weight_path = weight_path
        self._init_weight = init_weight
        self._softmax_at_test = softmax_at_test

        self.model = MMOVModel(
            model_path_or_model,
            weight_path,
            inputs=inputs,
            outputs=outputs,
            remove_normalize=False,
            merge_bn=False,
            paired_bn=False,
            init_weight=init_weight,
            verify_shape=verify_shape,
            parser=cls_base_parser,
            parser_kwargs=dict(component="head"),
        )

    def forward_train(self, x, gt_label, **kwargs):
        cls_score = self.model(x)
        while cls_score.dim() > 2:
            cls_score = cls_score.squeeze(2)
        losses = self.loss(cls_score, gt_label, **kwargs)
        return losses

    def simple_test(self, x):
        cls_score = self.model(x)
        while cls_score.dim() > 2:
            cls_score = cls_score.squeeze(2)
        if self._softmax_at_test:
            pred = F.softmax(cls_score, dim=1) if cls_score is not None else None
        else:
            pred = cls_score
        return self.post_process(pred)



"""Base MobileNet-V3."""

# model settings
model = dict(
    type="ImageClassifier",
    backbone=dict(type="otx.OTXMobileNetV3", pretrained=True, mode="small", width_mult=1.0),
    neck=dict(type="GlobalAveragePooling"),
    head=dict(
        type="NonLinearClsHead",
        num_classes=1000,
        in_channels=576,
        hid_channels=1024,
        act_cfg=dict(type="HSwish"),
        loss=dict(type="CrossEntropyLoss", loss_weight=1.0),
    ),
)
